from __future__ import annotations

import contextlib
import dataclasses
import io
import pathlib
import re
import sys
import traceback
from dataclasses import dataclass

import pytest

import yuio
import yuio.io
import yuio.term
import yuio.theme
import yuio.widget
from yuio import _typing as _t

T = _t.TypeVar("T")
W = _t.TypeVar("W", bound=yuio.widget.Widget[object])


_WIDTH = 20
_HEIGHT = 5


@pytest.fixture
def width() -> int:
    return _WIDTH


@pytest.fixture
def height() -> int:
    return _HEIGHT


@pytest.fixture
def ostream() -> io.StringIO:
    ostream = io.StringIO()
    ostream.isatty = lambda: True
    return ostream


@pytest.fixture
def istream() -> _t.TextIO:
    istream = _MockedIStream()
    istream.isatty = lambda: True
    return istream


@pytest.fixture
def theme() -> yuio.theme.Theme:
    return yuio.theme.Theme()


@pytest.fixture
def term(ostream: io.StringIO, istream: _t.TextIO) -> yuio.term.Term:
    return yuio.term.Term(
        ostream,
        istream,
        color_support=yuio.term.ColorSupport.ANSI_TRUE,
        interactive_support=yuio.term.InteractiveSupport.FULL,
    )


@pytest.fixture
def rc(
    term: yuio.term.Term, theme: yuio.theme.Theme, width: int, height: int
) -> yuio.widget.RenderContext:
    rc = yuio.widget.RenderContext(term, theme)
    rc._override_wh = (width, height)
    rc.prepare()
    return rc


@pytest.fixture
def io_mocker_factory(
    ostream: io.StringIO,
    term: yuio.term.Term,
    theme: yuio.theme.Theme,
    width: int,
    height: int,
) -> _t.Callable[[], IOMocker]:
    def factory():
        return IOMocker(ostream, term, theme, width, height)

    return factory


@pytest.fixture
def io_mocker(io_mocker_factory) -> IOMocker:
    return io_mocker_factory()


@pytest.fixture
def widget_checker_factory(
    ostream: io.StringIO,
    term: yuio.term.Term,
    theme: yuio.theme.Theme,
    width: int,
    height: int,
) -> _t.Callable[[], WidgetChecker[yuio.widget.Widget[object]]]:
    def factory():
        return WidgetChecker[yuio.widget.Widget[object]](
            ostream, term, theme, width, height
        )

    return factory


@pytest.fixture
def widget_checker(
    widget_checker_factory,
) -> WidgetChecker[yuio.widget.Widget[object]]:
    return widget_checker_factory()


@dataclass
class _KeyboardEventStreamDone(BaseException):
    def __str__(self):
        return self.__class__.__name__


_KeyboardEventStreamDone.__name__ = """<finish event stream>"""


@dataclass
class _ExpectStdinRead:
    result: str

    def __str__(self):
        return self.__class__.__name__


_ExpectStdinRead.__name__ = """<call to stdin.read>"""


@dataclass
class _ExpectStdinReadline:
    result: str

    def __str__(self):
        return self.__class__.__name__


_ExpectStdinReadline.__name__ = """<call to stdin.readline>"""


@dataclass
class _ExpectStdinReadlines:
    result: list[str]

    def __str__(self):
        return self.__class__.__name__


_ExpectStdinReadlines.__name__ = """<call to stdin.readlines>"""


@dataclass
class _WidgetAssert:
    fn: _t.Callable[[], None]

    def __str__(self):
        return self.__class__.__name__


_WidgetAssert.__name__ = """<assert widget state>"""


@dataclass
class _ExpectMark:
    mark: str | None

    def __str__(self):
        return f"<mark: {self.mark}>"


_ExpectMark.__name__ = """<mark>"""


class _KeyboardEventStream:
    def __init__(
        self,
        ostream: io.StringIO,
        term: yuio.term.Term,
        theme: yuio.theme.Theme,
        events: list[
            tuple[
                traceback.StackSummary,
                yuio.widget.KeyboardEvent
                | RcCompare
                | _WidgetAssert
                | _KeyboardEventStreamDone
                | _ExpectStdinRead
                | _ExpectStdinReadline
                | _ExpectStdinReadlines
                | _ExpectMark,
            ]
        ],
        width: int,
    ):
        self.ostream = ostream
        self.term = term
        self.theme = theme
        self.events = events
        self.index = 0
        self.width = width

    def __iter__(self):
        return self

    def __next__(self):
        return self._next()[1]

    def _next(
        self, expected_event_kind: type[T] = yuio.widget.KeyboardEvent
    ) -> tuple[traceback.StackSummary, T]:
        stack_summary, event = self._next_any()
        if not isinstance(event, expected_event_kind):
            raise AssertionError(
                f"expected {expected_event_kind.__name__}, got {event}"
            )
        return stack_summary, event

    def _next_any(self):
        while self.index < len(self.events):
            (stack_summary, event) = self.events[self.index]
            self.index += 1
            with self._patch_stack_summary(stack_summary):
                if isinstance(event, _KeyboardEventStreamDone):
                    raise event
                elif isinstance(event, RcCompare):
                    self.ostream.seek(0)
                    commands = self.ostream.read()
                    self.ostream.seek(0, io.SEEK_END)
                    assert RcCompare.from_commands(commands, self.width) == event
                elif isinstance(event, _WidgetAssert):
                    event.fn()
                else:
                    return stack_summary, event
        raise StopIteration

    def finish(self):
        try:
            # Process all asserts at the end of the stream.
            self._next_any()
        except StopIteration:
            pass
        else:
            # There are some unconsumed events.
            self.index -= 1  # Adjust for event that we've got from `_next_any`.
            raise AssertionError(
                "some events haven't been queried:\n\n"
                + "\n".join(f"{event[1]}" for event in self.events[self.index :])
            )

    def read(self, n: int = -1) -> str:
        try:
            _, event = self._next(expected_event_kind=_ExpectStdinRead)
        except StopIteration:
            raise AssertionError("unexpected call to istream.read()") from None
        self.ostream.write(event.result)
        return event.result

    def readline(self, limit: int = -1) -> str:
        try:
            _, event = self._next(expected_event_kind=_ExpectStdinReadline)
        except StopIteration:
            raise AssertionError("unexpected call to istream.readline()") from None
        self.ostream.write(event.result)
        return event.result

    def readlines(self, hint: int = -1) -> list[str]:
        try:
            _, event = self._next(expected_event_kind=_ExpectStdinReadlines)
        except StopIteration:
            raise AssertionError("unexpected call to istream.read()") from None
        self.ostream.writelines(event.result)
        return event.result

    def mark(self, mark: str | None):
        try:
            stack_summary, event = self._next(expected_event_kind=_ExpectMark)
        except StopIteration:
            raise AssertionError("unexpected call to io_mocker.mark()") from None
        if event.mark != mark:
            with self._patch_stack_summary(stack_summary):
                raise AssertionError(f"expected {event}, got {_ExpectMark(mark)}")

    @contextlib.contextmanager
    def _patch_stack_summary(self, stack_summary: traceback.StackSummary):
        try:
            yield
        except Exception as e:
            stack_summary_text = "".join(stack_summary.format())
            e.args = (f"{e}\n\nAssertion origin:\n\n{stack_summary_text}",)
            raise e


_CURRENT_IOSTREAM_MOCK: _KeyboardEventStream | None = None


class IOMocker:
    """
    A class for mocking contents of input/output streams and keyboard event stream.

    All tests use a separate terminal (an instance of :class:`yuio.term.Term`),
    that's been initialized with mock streams (specifically, :class:`io.StringIO`
    as output, and :class:`_MockedIStream` as input). This class can manipulate these
    streams, as well as :func:`yuio.widget._event_stream`, so that they yield
    what we expect.

    See fixtures for more details.

    """

    def __init__(
        self,
        ostream: io.StringIO | None = None,
        term: yuio.term.Term | None = None,
        theme: yuio.theme.Theme | None = None,
        width: int | None = None,
        height: int | None = None,
        wrap_streams: bool = False,
    ):
        self._events: list[
            tuple[
                traceback.StackSummary,
                yuio.widget.KeyboardEvent
                | RcCompare
                | _WidgetAssert
                | _KeyboardEventStreamDone
                | _ExpectStdinRead
                | _ExpectStdinReadline
                | _ExpectStdinReadlines
                | _ExpectMark,
            ]
        ] = []

        self._ostream = ostream
        self._term = term
        self._theme = theme
        self._width = width
        self._height = height
        self._wrap_streams = wrap_streams

    def refresh(self) -> _t.Self:
        """
        Refresh screen.

        """
        return self.key("l", ctrl=True)

    def key(
        self,
        key: str | yuio.widget.Key,
        ctrl: bool = False,
        alt: bool = False,
    ) -> _t.Self:
        """
        Yield a key from the mocked keyboard event stream.

        """
        self._events.append(
            (self._get_stack_summary(), yuio.widget.KeyboardEvent(key, ctrl, alt))
        )
        return self

    def keyboard_event(self, keyboard_event: yuio.widget.KeyboardEvent) -> _t.Self:
        """
        Yield a keyboard event from the mocked keyboard event stream.

        """
        self._events.append((self._get_stack_summary(), keyboard_event))
        return self

    def text(self, text: str) -> _t.Self:
        """
        Yield a text, one letter at a time, from the mocked keyboard event stream.

        """
        stack_summary = self._get_stack_summary()
        self._events.extend((stack_summary, yuio.widget.KeyboardEvent(c)) for c in text)
        return self

    def paste(self, text: str) -> _t.Self:
        """
        Yield a text as a single paste event, from the mocked keyboard event stream.

        """
        stack_summary = self._get_stack_summary()
        self._events.append(
            (
                stack_summary,
                yuio.widget.KeyboardEvent(yuio.widget.Key.PASTE, paste_str=text),
            )
        )
        return self

    def expect_screen(
        self,
        screen: list[str] | None = None,
        colors: list[str] | None = None,
        cursor_x: int | None = None,
        cursor_y: int | None = None,
    ) -> _t.Self:
        """
        Check that the current contents of the output stream,
        when rendered onto a terminal, produce the given screen contents.

        """
        self._events.append(
            (self._get_stack_summary(), RcCompare(screen, colors, cursor_x, cursor_y))
        )
        return self

    def expect(self, fn: _t.Callable[[], bool]) -> _t.Self:
        """
        Run a lambda function and assert that it returns :data:`True`.

        """

        def cb():
            assert fn()

        self._events.append((self._get_stack_summary(), _WidgetAssert(cb)))
        return self

    def expect_eq(self, fn: _t.Callable[[], T], expected: T) -> _t.Self:
        """
        Run a lambda function and assert that it returns a value
        equal to the `expected`.

        """

        def cb():
            assert fn() == expected

        self._events.append((self._get_stack_summary(), _WidgetAssert(cb)))
        return self

    def expect_ne(self, fn: _t.Callable[[], T], expected: T) -> _t.Self:
        """
        Run a lambda function and assert that it returns a value
        not equal to the `expected`.

        """

        def cb():
            assert fn() != expected

        self._events.append((self._get_stack_summary(), _WidgetAssert(cb)))
        return self

    def expect_widget_to_continue(self) -> _t.Self:
        """
        Expect that a widget requests another event from
        the mocked keyboard event stream.

        This function will terminate mocked block early by raising an exception.
        The :meth:`KeyboardEventStream.mock` will catch it.

        """
        self._events.append((self._get_stack_summary(), _KeyboardEventStreamDone()))
        return self

    def expect_istream_read(self, result: str) -> _t.Self:
        """
        Expect a call to `istream.read()` and return the given result from it.

        """
        self._events.append((self._get_stack_summary(), _ExpectStdinRead(result)))
        return self

    def expect_istream_readline(self, result: str) -> _t.Self:
        """
        Expect a call to `istream.readline()` and return the given result from it.

        """
        self._events.append((self._get_stack_summary(), _ExpectStdinReadline(result)))
        return self

    def expect_istream_read(self, result: list[str]) -> _t.Self:
        """
        Expect a call to `istream.readlines()` and return the given result from it.

        """
        self._events.append((self._get_stack_summary(), _ExpectStdinReadlines(result)))
        return self

    def expect_mark(self, mark: str | None = None) -> _t.Self:
        """
        Expect a call to :meth:`IoMocker.mark`.

        This method allows adding custom checkpoints to the test code.
        This is useful when testing code that renders something but doesn't request
        any input from a user. Without a mark, all asserts will be checked at the end
        of the test, when :meth:`IoMocker.mock` exits. With a mark, you can manually
        trigger assert checks between redraws.

        Example:

        .. code-block:: python

           def test_something(io_mocker):
              # First assert will be checked upon a call to `mark`.
              io_mocker.expect_screen(...)
              io_mocker.expect_mark()

              # Second assert will be checked at the end of the test.
              io_mocker.expect_screen(...)

              with io_mocker.mock():
                  # Render things
                  ...

                  # Emit a mark event; this will check what we've rendered so far.
                  io_mocker.mark()

                  # Render some more things
                  ...

        """
        self._events.append((self._get_stack_summary(), _ExpectMark(mark)))
        return self

    def mark(self, mark: str | None = None):
        """
        Trigger a mark event.

        See :meth:`IoMocker.expect_mark` for details.

        """
        if _CURRENT_IOSTREAM_MOCK is None:
            raise RuntimeError(
                "you need to mock io streams before using this function; "
                "call KeyboardEventStream.mock or WidgetChecker.check to do so"
            )

        _CURRENT_IOSTREAM_MOCK.mark(mark)

    @staticmethod
    def _get_stack_summary() -> traceback.StackSummary:
        base_path = str(pathlib.Path(__file__).parent.parent.absolute())
        stack_summary = traceback.extract_stack()
        stack_summary.pop()
        stack_summary.pop()
        while stack_summary and base_path not in stack_summary[0].filename:
            stack_summary.pop(0)
        return stack_summary

    @contextlib.contextmanager
    def mock(
        self,
        ostream: io.StringIO | None = None,
        term: yuio.term.Term | None = None,
        theme: yuio.theme.Theme | None = None,
        width: int | None = None,
        height: int | None = None,
        wrap_streams: bool | None = None,
    ):
        """
        Bind :func:`yuio.widget._event_stream` and all mocked inout streams
        to events from this mocker.

        """

        ostream = ostream or self._ostream
        if ostream is None:
            raise RuntimeError(
                "this mocker is not bound to any ostream; "
                "pass ostream to its constructor or its `mock`/`check` method"
            )

        term = term or self._term
        if term is None:
            raise RuntimeError(
                "this mocker is not bound to any term; "
                "pass term to its constructor or its `mock`/`check` method"
            )

        theme = theme or self._theme
        if theme is None:
            raise RuntimeError(
                "this mocker is not bound to any theme; "
                "pass theme to its constructor or its `mock`/`check` method"
            )

        width = width or self._width
        if width is None:
            raise RuntimeError(
                "this mocker is not bound to any width; "
                "pass width to its constructor or its `mock`/`check` method"
            )

        height = height or self._height
        if height is None:
            raise RuntimeError(
                "this mocker is not bound to any height; "
                "pass height to its constructor or its `mock`/`check` method"
            )

        wrap_streams = wrap_streams if wrap_streams is not None else self._wrap_streams

        global _CURRENT_IOSTREAM_MOCK
        if _CURRENT_IOSTREAM_MOCK is not None:
            raise RuntimeError("can't have more than one mock at a time")
        _CURRENT_IOSTREAM_MOCK = _KeyboardEventStream(
            ostream, term, theme, self._events, width
        )

        old_event_stream, yuio.widget._event_stream = (
            yuio.widget._event_stream,
            lambda: _CURRENT_IOSTREAM_MOCK,
        )
        old_enter_raw_mode, yuio.term._enter_raw_mode = (
            yuio.term._enter_raw_mode,
            lambda *_, **__: contextlib.nullcontext(),
        )
        old_stderr, sys.stderr = sys.stderr, term.ostream
        old_stdout, sys.stdout = sys.stdout, term.ostream
        old_stdin, sys.stdin = sys.stdin, term.istream

        yuio.widget.RenderContext._override_wh = (width, height)

        try:
            if wrap_streams:
                assert not yuio.io.streams_wrapped(), "previous test didn't clean up?"
                yuio.io.wrap_streams()

            try:
                yield
            except _KeyboardEventStreamDone:
                _CURRENT_IOSTREAM_MOCK.finish()
            else:
                _CURRENT_IOSTREAM_MOCK.finish()
        finally:
            if wrap_streams:
                yuio.io.restore_streams()
            yuio.widget.RenderContext._override_wh = None
            sys.stdin = old_stdin
            sys.stdout = old_stdout
            sys.stderr = old_stderr
            yuio.term._enter_raw_mode = old_enter_raw_mode
            yuio.widget._event_stream = old_event_stream
            _CURRENT_IOSTREAM_MOCK = None


class WidgetChecker(IOMocker, _t.Generic[W]):
    """
    A keyboard event stream mocker that checks a specific widget.

    This class works like `KeyboardEventStream`, but it also knows that we're testing
    a specific widget. Because of that, we're able to write asserts
    on widget properties. For example:

    .. code-block:: python

       def test_something(ostream, term, theme, width, height):
          # We know that we're testing an `Input` widget...
          checker = (
              WidgetChecker[yuio.widget.Input](ostream, term, theme, width, height)
                  # ...therefore we can write asserts that check specific properties
                  # of the `Input` widget.
                  .expect_widget_eq(lambda widget: widget.text, "foo bar!")
                  .expect_widget_to_continue()
          )

          # Now we can run tests for a widget.
          checker.check(yuio.widget.Input(text="foo bar!"))

    .. invisible-code-block: python

       _ostream = io.StringIO()
       _istream = _MockedIStream()
       _term = yuio.term.Term(
           _ostream,
           _istream,
           color_support=yuio.term.ColorSupport.ANSI_TRUE,
           interactive_support=yuio.term.InteractiveSupport.FULL,
       )
       _theme = yuio.theme.Theme()
       test_something(_ostream, _term, _theme, 20, 5)

    """

    _widget: W | None = None

    def call(self, fn: _t.Callable[[W], None]) -> _t.Self:
        """
        Run a function with widget as an input.

        """

        def widget_fn():
            assert self._widget is not None, "widget is required"
            fn(self._widget)

        self._events.append((self._get_stack_summary(), _WidgetAssert(widget_fn)))
        return self

    def expect_widget(self, fn: _t.Callable[[W], bool]) -> _t.Self:
        """
        Assert some property of a widget.

        """

        def widget_fn():
            assert self._widget is not None, "widget is required"
            assert fn(self._widget)

        self._events.append((self._get_stack_summary(), _WidgetAssert(widget_fn)))
        return self

    def expect_widget_eq(self, fn: _t.Callable[[W], T], expected: T) -> _t.Self:
        """
        Assert that some property of a widget equals to the given value.

        """

        def widget_fn():
            assert self._widget is not None, "widget is required"
            assert fn(self._widget) == expected

        self._events.append((self._get_stack_summary(), _WidgetAssert(widget_fn)))
        return self

    def expect_widget_ne(self, fn: _t.Callable[[W], T], expected: T) -> _t.Self:
        """
        Assert that some property of a widget is not equal to the given value.

        """

        def widget_fn():
            assert self._widget is not None, "widget is required"
            assert fn(self._widget) != expected

        self._events.append((self._get_stack_summary(), _WidgetAssert(widget_fn)))
        return self

    @contextlib.contextmanager
    def mock(
        self,
        ostream: io.StringIO | None = None,
        term: yuio.term.Term | None = None,
        theme: yuio.theme.Theme | None = None,
        width: int | None = None,
        height: int | None = None,
        wrap_streams: bool | None = None,
    ):
        raise RuntimeError("use WidgetChecker.check() instead")

    def check(
        self,
        widget: W,
        ostream: io.StringIO | None = None,
        term: yuio.term.Term | None = None,
        theme: yuio.theme.Theme | None = None,
        width: int | None = None,
        height: int | None = None,
        wrap_streams: bool | None = None,
    ):
        """
        Run tests on the given widget, and return widget's result.

        """

        assert self._widget is None, "can't have more than one widget check at a time"
        self._widget = widget
        with super().mock(ostream, term, theme, width, height, wrap_streams):
            if _CURRENT_IOSTREAM_MOCK is None:
                raise RuntimeError(
                    "mock() should've set _CURRENT_IOSTREAM_MOCK, but it didn't?"
                )

            try:
                return widget.run(
                    _CURRENT_IOSTREAM_MOCK.term, _CURRENT_IOSTREAM_MOCK.theme
                )
            finally:
                self._widget = None


class _MockedIStream(_t.TextIO):
    _closed: bool = False

    def close(self):
        self._closed = True

    def _assert_not_closed(self):
        if self._closed:
            raise ValueError("I/O operation on closed file")

    @property
    def closed(self) -> bool:
        return self._closed

    def fileno(self) -> int:
        raise io.UnsupportedOperation("fileno")

    def flush(self):
        self._assert_not_closed()

    def isatty(self) -> bool:
        return False

    def writable(self) -> bool:
        return False

    def write(self, s: str, /) -> int:
        raise io.UnsupportedOperation("write")

    def writelines(self, lines: _t.Iterable[str], /):
        raise io.UnsupportedOperation("writelines")

    def readable(self) -> bool:
        return True

    def read(self, n: int = -1) -> str:
        self._assert_not_closed()
        if _CURRENT_IOSTREAM_MOCK is None:
            raise RuntimeError(
                "you need to mock io streams before using this function; "
                "call KeyboardEventStream.mock or WidgetChecker.check to do so"
            )
        return _CURRENT_IOSTREAM_MOCK.read(n)

    def readline(self, limit: int = -1) -> str:
        self._assert_not_closed()
        if _CURRENT_IOSTREAM_MOCK is None:
            raise RuntimeError(
                "you need to mock io streams before using this function; "
                "call KeyboardEventStream.mock or WidgetChecker.check to do so"
            )
        return _CURRENT_IOSTREAM_MOCK.readline(limit)

    def readlines(self, hint: int = -1) -> list[str]:
        self._assert_not_closed()
        if _CURRENT_IOSTREAM_MOCK is None:
            raise RuntimeError(
                "you need to mock io streams before using this function; "
                "call KeyboardEventStream.mock or WidgetChecker.check to do so"
            )
        return _CURRENT_IOSTREAM_MOCK.readlines(hint)

    def seek(self, offset: int, whence: int = 0) -> int:
        raise io.UnsupportedOperation("underlying stream is not seekable")

    def seekable(self) -> bool:
        return False

    def tell(self) -> int:
        raise io.UnsupportedOperation("underlying stream is not seekable")

    def truncate(self, size: int | None = None) -> int:
        raise io.UnsupportedOperation("truncate")

    def __enter__(self) -> _t.TextIO:
        self._assert_not_closed()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def encoding(self) -> str:
        return None  # type: ignore

    @property
    def errors(self) -> str | None:
        return None


def pytest_assertrepr_compare(op, left, right):
    if op == "==" and (isinstance(left, RcCompare) or isinstance(right, RcCompare)):
        if isinstance(left, str):
            left = RcCompare.from_commands(left)
        elif not isinstance(left, RcCompare):
            return None
        if isinstance(right, str):
            right = RcCompare.from_commands(right)
        elif not isinstance(right, RcCompare):
            return None
        return _rc_diff(left, right)


@dataclass
class RcCompare:
    screen: list[str] | None = None
    colors: list[str] | None = None
    cursor_x: int | None = None
    cursor_y: int | None = None

    commands: str = dataclasses.field(default="", compare=False, hash=False)

    def __post_init__(self, *args, **kwargs):
        if self.screen is not None and self.colors is not None:
            if len(self.screen) != len(self.colors):
                raise RuntimeError(
                    f"screen height does not match colors height "
                    f"({len(self.screen)} != {len(self.colors)})"
                )

        screen_width = None
        if self.screen:
            screen_width = yuio.term.line_width(self.screen[0])
            for i, line in enumerate(self.screen):
                if yuio.term.line_width(line) != screen_width:
                    raise RuntimeError(
                        f"width of screen line {i + 1} is not equal to the width of screen line 1: "
                        f"({yuio.term.line_width(self.screen[i])} != {screen_width})"
                    )

        colors_width = None
        if self.colors:
            colors_width = yuio.term.line_width(self.colors[0])
            for i, line in enumerate(self.colors):
                if yuio.term.line_width(line) != colors_width:
                    raise RuntimeError(
                        f"width of colors line {i + 1} is not equal to the width of colors line 1: "
                        f"({yuio.term.line_width(self.colors[i])} != {colors_width})"
                    )

        if screen_width is not None and colors_width is not None:
            if screen_width != colors_width:
                raise RuntimeError(
                    f"screen width does not match colors width "
                    f"({screen_width} != {colors_width})"
                )

        while self.screen or self.colors:
            screen_is_space = self.screen is None or self.screen[-1].isspace()
            colors_is_space = self.colors is None or self.colors[-1].isspace()
            if screen_is_space and colors_is_space:
                if self.screen:
                    self.screen.pop()
                if self.colors:
                    self.colors.pop()
            elif not screen_is_space or not colors_is_space:
                break

    @classmethod
    def from_commands(cls, commands: str, width: int = _WIDTH):
        return cls(*_render_screen(commands, width), commands)

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, RcCompare):
            return NotImplemented

        if None not in [self.screen, rhs.screen] and self.screen != rhs.screen:
            return False
        if None not in [self.colors, rhs.colors] and self.colors != rhs.colors:
            return False
        if None not in [self.cursor_x, rhs.cursor_x] and self.cursor_x != rhs.cursor_x:
            return False
        if None not in [self.cursor_y, rhs.cursor_y] and self.cursor_y != rhs.cursor_y:
            return False

        return True

    def __str__(self):
        return "<assert rendered screen>"


_CSI_RE = re.compile(r"\x1b\[((?:-?[0-9]+)?(?:;(?:-?[0-9]+)?)*(?:[mJHABCDG]))|\a")
_COLOR_NAMES = "krgybmcw"


def _rc_diff(a: RcCompare, b: RcCompare):
    out = ["Comparing rendering results"]
    if a.screen is not None and b.screen is not None:
        out += _show_diff(a.screen, b.screen, "Text:")
    if a.colors is not None and b.colors is not None:
        out += _show_diff(a.colors, b.colors, "Colors:")
    out.append("")
    if None not in [a.cursor_x, a.cursor_y]:
        out.append(f"Left cursor: ({a.cursor_x}, {a.cursor_y})")
    if None not in [b.cursor_x, b.cursor_y]:
        out.append(f"Right cursor: ({b.cursor_x}, {b.cursor_y})")
    if a.commands:
        out.append(f"Left commands: {_sanitize_commands(a.commands)}")
    if b.commands:
        out.append(f"Left commands: {_sanitize_commands(b.commands)}")
    return out


def _show_diff(a_screen: list[str], b_screen: list[str], what: str) -> list[str]:
    if a_screen == b_screen:
        return []

    a_height = len(a_screen)
    b_height = len(b_screen)

    if not a_height and not b_height:
        return []

    a_width = len(a_screen[0]) if a_screen else 0
    b_width = len(b_screen[0]) if b_screen else 0
    if not a_screen:
        a_width = b_width
    elif not b_screen:
        b_width = a_width
    if a_width != b_width:
        return [f"Screen widths differ: {a_width} != {b_width}"]
    width = max(a_width, b_width)

    if not width:
        return []

    out_h = "   expected"
    out_expected = [
        out_h + " " * (width + 5 - len(out_h)),
        "   ┌" + "─" * width + "┐",
        *[f"{i + 1: >2} │{line}│" for i, line in enumerate(b_screen)],
        "   └" + "─" * width + "┘",
    ]

    got_h = "   actual"
    out_got = [
        got_h + " " * (width + 5 - len(got_h)),
        "   ┌" + "─" * width + "┐",
        *[f"{i + 1: >2} │{line}│" for i, line in enumerate(a_screen)],
        "   └" + "─" * width + "┘",
    ]

    diff_h = "   diff"
    out_diff = [
        diff_h + " " * (width + 5 - len(diff_h)),
        "   ┌" + "─" * width + "┐",
    ]

    for i in range(max(a_height, b_height)):
        line = f"{i + 1: >2} │"
        a_line = a_screen[i] if i < a_height else "\0" * width
        b_line = b_screen[i] if i < b_height else "\0" * width
        for a, b in zip(a_line, b_line):
            line += " " if a == b else "!"
        out_diff.append(line + "│")

    out_diff.append("   └" + "─" * width + "┘")

    height = max(len(out_got), len(out_expected), len(out_diff))

    out_got += ["     " + " " * width for _ in range(height - len(out_got))]
    out_expected += ["     " + " " * width for _ in range(height - len(out_expected))]
    out_diff += ["     " + " " * width for _ in range(height - len(out_diff))]

    return [what] + [
        f"{expected_line} {got_line} {diff_line}"
        for expected_line, got_line, diff_line in zip(out_expected, out_got, out_diff)
    ]


def _render_screen(commands: str, width: int) -> tuple[list[str], list[str], int, int]:
    height = 0
    x, y = 0, 0
    text = []
    colors = []
    color = " "

    for i, part in enumerate(_CSI_RE.split(commands)):
        if i % 2 == 0:
            # Render text.
            for c in part:
                assert c not in "\r"
                if c == "\n":
                    x = 0
                    y += 1
                else:
                    cw = yuio.term.line_width(c)
                    if cw == 0:
                        raise RuntimeError("this checker can't handle zero-width chars")

                    if x < 0 or y < 0:
                        raise RuntimeError(
                            f"printing at negative coordinates: ({x}, {y})"
                        )
                    if x + cw > width:
                        x = 0
                        y += 1
                    if y >= height:
                        text += [[" "] * width for _ in range(y - height + 1)]
                        colors += [[" "] * width for _ in range(y - height + 1)]
                        height = y + 1
                    for _ in range(cw):
                        text[y][x] = c
                        colors[y][x] = color
                        c = ""
                        x += 1
        else:
            # Render a CSI.

            if not part:
                continue  # '\a'

            fn = part[-1]
            args = part[:-1].split(";")

            if fn == "m":
                # Color.
                for code in part[:-1].split(";"):
                    bold = False
                    if not code or code in ["0", "39"]:
                        color = " "
                    else:
                        int_code = int(code)
                        if int_code == 1:
                            bold = True
                        elif int_code == 2:
                            pass  # dim
                        elif 30 <= int_code <= 37:
                            color = _COLOR_NAMES[int_code - 30]
                        else:
                            assert (
                                False
                            ), f"don't use non-standard colors with this assertion: {int_code}"
                    if bold:
                        color = "#" if color == " " else color.upper()
            elif fn == "J":
                # Clear screen.
                assert len(args) <= 1, f"invalid OSC: {part!r}"
                if not args or args[0] in ("", "0"):
                    del text[y:]
                    del colors[y:]
                    height = y
                elif args[0] == "2":
                    text = []
                    colors = []
                    height = 0
                else:
                    assert False, f"invalid OSC: {part!r}"
            elif fn == "H":
                # Absolute cursor position.
                if len(args) == 0:
                    y, x = 0, 0
                elif len(args) == 1:
                    y, x = int(args[0] or "1") - 1, 0
                elif len(args) == 2:
                    y, x = int(args[0] or "1") - 1, int(args[1] or "1") - 1
                else:
                    assert False, f"invalid OSC: {part!r}"
            elif fn == "A":
                # Cursor up.
                assert len(args) <= 1, f"invalid OSC: {part!r}"
                y -= int(args[0] or "1") if args else 1
            elif fn == "B":
                # Cursor down.
                assert len(args) <= 1, f"invalid OSC: {part!r}"
                y += int(args[0] or "1") if args else 1
            elif fn == "C":
                # Cursor forward.
                assert len(args) <= 1, f"invalid OSC: {part!r}"
                x += int(args[0] or "1") if args else 1
            elif fn == "D":
                # Cursor back.
                assert len(args) <= 1, f"invalid OSC: {part!r}"
                x -= int(args[0] or "1") if args else 1
            elif fn == "G":
                # Absolute horizontal cursor position.
                assert len(args) <= 1, f"invalid OSC: {part!r}"
                x = int(args[0] or "1") - 1 if args else 1

    return (
        ["".join(line) for line in text],
        ["".join(line) for line in colors],
        x,
        y,
    )


def _sanitize_commands(commands: str) -> str:
    return (
        commands.replace("\x1b", "ESC")
        .replace("\a", "BEL")
        .replace("\r", "CR")
        .replace("\n", "LF")
    )



import yuio.io

TEXT = """# Please edit the lines below. Lines starting with '#' will be ignored.

Sing, O goddess, the anger of Achilles son of Peleus,
that brought countless ills upon the Achaeans.
Many a brave soul did it send hurrying down to Hades,
and many a hero did it yield a prey to dogs and vultures,
for so were the counsels of Jove fulfilled from the day on which the son of Atreus,
king of men, and great Achilles, first fell out with one another.
"""

if __name__ == "__main__":
    yuio.io.heading("Yuio's interactive editing showcase")

    yuio.io.info(
        "This functionality is similar to what GIT does when you commit something:"
    )
    yuio.io.info("it opens your default editor and lets you edit a commit message.")

    yuio.io.br()

    yuio.io.wait_for_user()

    result = yuio.io.edit(TEXT)

    yuio.io.br()

    yuio.io.success("Editing successful!")
    yuio.io.success("So, this is what you've done to Homer's Iliad:")

    yuio.io.br()

    if result:
        yuio.io.info(result)
    else:
        yuio.io.info("...")
        yuio.io.br()
        yuio.io.info("... nothing! You've deleted it! How could you 〈◕﹏◕〉")



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
Everything to do with terminal output: detecting terminal capabilities,
working with colors and themes, formatting text.

This is a low-level API module, upon which :mod:`yuio.io` builds
its higher-level abstraction.


Detecting terminal capabilities
-------------------------------

Terminal capabilities are stored in a :class:`Term` object.

Usually, you don't need to query terminal capabilities yourself,
as you can use Yuio's global configuration from :mod:`yuio.io`
(see :func:`yuio.io.get_term`).

However, you can get a :class:`Term` object by using :func:`get_term_from_stream`:

.. autofunction:: get_term_from_stream

:class:`Term` contains all info about what kinds of things the terminal
supports. If available, it will also have info about terminal's theme,
i.e. dark or light background, etc:

.. autoclass:: Term
   :members:

.. autoclass:: Lightness
   :members:

.. autoclass:: ColorSupport
   :members:

.. autoclass:: InteractiveSupport
   :members:

.. autoclass:: TerminalColors
   :members:


Working with colors
-------------------

Text background and foreground color, as well as its style, is defined
by the :class:`Color` class. It stores RGB components and ANSI escape codes
for every aspect of text presentation:

.. autoclass:: Color
   :members:

A single color value is stored in the :class:`ColorValue` class.
Usually you don't need to use these directly, as you will be working
with :class:`Color` instances. Nevertheless, you can operate individual
colors should you need such thing:

.. autoclass:: ColorValue
   :members:


Coloring and formatting text
----------------------------

The higher-level :mod:`io` module uses strings with xml-like color
tags to transfer information about line formatting. Here, on the lower level,
these strings are parsed and transformed into :class:`ColorizedString`:

.. autoclass:: ColorizedString
   :members:

.. class:: RawString

    Raw colorized string. This is the underlying type
    for :class:`ColorizedString`.

.. class:: AnyString

    Any string (i.e. a :class:`str`, a raw colorized string,
    or a normal colorized string).

.. autoclass:: NoWrap

.. autoclass:: Esc


Utilities
---------

.. autofunction:: line_width

"""

from __future__ import annotations

import colorsys
import contextlib
import dataclasses
import enum
import functools
import os
import re
import sys
import typing
import unicodedata
from dataclasses import dataclass

import yuio
from yuio import _typing as _t

__all__ = [
    "Lightness",
    "ColorSupport",
    "InteractiveSupport",
    "TerminalColors",
    "Term",
    "get_term_from_stream",
    "ColorValue",
    "Color",
    "line_width",
    "NoWrap",
    "Esc",
    "ColorizedString",
    "RawString",
    "AnyString",
]

T = _t.TypeVar("T")


class Lightness(enum.Enum):
    """
    Overall color theme of a terminal.

    Can help with deciding which colors to use when printing output.

    """

    UNKNOWN = enum.auto()
    """
    We couldn't determine terminal background, or it wasn't dark
    or bright enough to fall in one category or another.

    """

    DARK = enum.auto()
    """
    Terminal background is dark.

    """

    LIGHT = enum.auto()
    """
    Terminal background is light.

    """


class ColorSupport(enum.IntEnum):
    """
    Terminal's capability for coloring output.

    """

    NONE = 0
    """
    Color codes are not supported.

    """

    ANSI = 1
    """
    Only simple 8-bit color codes are supported.

    """

    #: 256-encoded colors are supported.
    ANSI_256 = 2

    ANSI_TRUE = 3
    """
    True colors are supported.

    """


class InteractiveSupport(enum.IntEnum):
    """
    Terminal's capability for rendering interactive widgets.

    """

    NONE = 0
    """
    Terminal can't render anything interactive.

    """

    MOVE_CURSOR = 1
    """
    Terminal can move cursor and erase lines.

    """

    FULL = 2
    """
    Terminal can process queries, enter ``CBREAK`` mode, etc.

    """


@dataclass(frozen=True)
class TerminalColors:
    """
    Colors and theme of the attached terminal.

    """

    background: ColorValue
    """
    Background color of a terminal.

    """

    foreground: ColorValue
    """
    Foreground color of a terminal.

    """

    black: ColorValue
    """
    Color value for the default "black" color.

    """

    red: ColorValue
    """
    Color value for the default "red" color.

    """

    green: ColorValue
    """
    Color value for the default "green" color.

    """

    yellow: ColorValue
    """
    Color value for the default "yellow" color.

    """

    blue: ColorValue
    """
    Color value for the default "blue" color.

    """

    magenta: ColorValue
    """
    Color value for the default "magenta" color.

    """

    cyan: ColorValue
    """
    Color value for the default "cyan" color.

    """

    white: ColorValue
    """
    Color value for the default "white" color.

    """

    lightness: Lightness
    """
    Overall color theme of a terminal, i.e. dark or light.

    """


@dataclass(frozen=True)
class Term:
    """
    Overall info about a terminal.

    """

    ostream: _t.TextIO
    """
    Terminal's output stream.

    """

    istream: _t.TextIO
    """
    Terminal's input stream.

    """

    color_support: ColorSupport = ColorSupport.NONE
    """
    Terminal's capability for coloring output.

    """

    interactive_support: InteractiveSupport = InteractiveSupport.NONE
    """
    Terminal's capability for rendering interactive widgets.

    """

    terminal_colors: TerminalColors | None = None
    """
    Terminal color settings.

    """

    @property
    def has_colors(self) -> bool:
        """
        Return :data:`True` if terminal supports simple 8-bit color codes.

        """

        return self.color_support >= ColorSupport.ANSI

    @property
    def has_colors_256(self) -> bool:
        """
        Return :data:`True` if terminal supports 256-encoded colors.

        """

        return self.color_support >= ColorSupport.ANSI_256

    @property
    def has_colors_true(self) -> bool:
        """
        Return :data:`True` if terminal supports true colors.

        """

        return self.color_support >= ColorSupport.ANSI_TRUE

    @property
    def can_move_cursor(self) -> bool:
        """
        Return :data:`True` if terminal can move cursor and erase lines.

        """

        return (
            self.has_colors
            and self.interactive_support >= InteractiveSupport.MOVE_CURSOR
        )

    @property
    def can_query_terminal(self) -> bool:
        """
        Return :data:`True` if terminal can process queries, enter CBREAK mode, etc.

        This is an alias to :attr:`~Term.is_fully_interactive`.

        """

        return self.is_fully_interactive

    @property
    def is_fully_interactive(self) -> bool:
        """
        Return :data:`True` if we're in a fully interactive environment.

        """

        return self.has_colors and self.interactive_support >= InteractiveSupport.FULL


_CI_ENV_VARS = [
    "TRAVIS",
    "CIRCLECI",
    "APPVEYOR",
    "GITLAB_CI",
    "BUILDKITE",
    "DRONE",
    "TEAMCITY_VERSION",
]


def get_term_from_stream(
    ostream: _t.TextIO, istream: _t.TextIO, /, *, query_terminal_colors: bool = True
) -> Term:
    """
    Query info about a terminal attached to the given stream.

    If ``query_terminal_colors`` is :data:`True`, Yuio will try to query background
    and foreground color of the terminal.

    """

    if "__YUIO_FORCE_FULL_TERM_SUPPORT" in os.environ:
        # For building docs in github
        return Term(
            ostream=ostream,
            istream=istream,
            color_support=ColorSupport.ANSI_TRUE,
            interactive_support=InteractiveSupport.FULL,
            terminal_colors=_get_standard_colors(ostream),
        )

    # Note: we don't rely on argparse to parse out flags and send them to us
    # because these functions can be called before parsing arguments.
    if (
        "--no-color" in sys.argv
        or "--no-colors" in sys.argv
        or "--force-no-color" in sys.argv
        or "--force-no-colors" in sys.argv
        or "FORCE_NO_COLOR" in os.environ
        or "FORCE_NO_COLORS" in os.environ
    ):
        return Term(ostream, istream)

    term = os.environ.get("TERM", "").lower()
    colorterm = os.environ.get("COLORTERM", "").lower()

    has_interactive_output = _is_interactive_output(ostream)
    has_interactive_input = _is_interactive_input(istream)
    is_foreground = _is_foreground(ostream) and _is_foreground(istream)

    color_support = ColorSupport.NONE
    in_ci = "CI" in os.environ
    if (
        "--force-color" in sys.argv
        or "--force-colors" in sys.argv
        or "FORCE_COLOR" in os.environ
        or "FORCE_COLORS" in os.environ
    ):
        color_support = ColorSupport.ANSI
    if has_interactive_output:
        if os.name == "nt":
            if _enable_vt_processing(ostream):
                color_support = ColorSupport.ANSI_TRUE
        elif "GITHUB_ACTIONS" in os.environ:
            color_support = ColorSupport.ANSI_TRUE
            in_ci = True
        elif any(ci in os.environ for ci in _CI_ENV_VARS):
            color_support = ColorSupport.ANSI
            in_ci = True
        elif colorterm in ("truecolor", "24bit") or term == "xterm-kitty":
            color_support = ColorSupport.ANSI_TRUE
        elif colorterm in ("yes", "true") or "256color" in term or term == "screen":
            color_support = ColorSupport.ANSI_256
        elif "linux" in term or "color" in term or "ansi" in term or "xterm" in term:
            color_support = ColorSupport.ANSI

    interactive_support = InteractiveSupport.NONE
    theme = None
    if is_foreground and color_support >= ColorSupport.ANSI and not in_ci:
        if has_interactive_output and has_interactive_input:
            interactive_support = InteractiveSupport.FULL
            if (
                query_terminal_colors
                and color_support >= ColorSupport.ANSI_256
                and "YUIO_DISABLE_OSC_QUERIES" not in os.environ
            ):
                theme = _get_standard_colors(ostream)
        else:
            interactive_support = InteractiveSupport.MOVE_CURSOR

    return Term(
        ostream=ostream,
        istream=istream,
        color_support=color_support,
        interactive_support=interactive_support,
        terminal_colors=theme,
    )


def _get_standard_colors(stream: _t.TextIO) -> TerminalColors | None:
    try:
        query = "\x1b]10;?\x1b\\\x1b]11;?\x1b\\" + "".join(
            [f"\x1b]4;{i};?\x1b\\" for i in range(8)]
        )
        response = _query_term(stream, query)
        if not response:
            return None

        # Deal with foreground color.

        match = re.match(
            r"^\x1b]10;rgb:([0-9a-f]{2,4})/([0-9a-f]{2,4})/([0-9a-f]{2,4})(?:\x1b\\|\a)",
            response,
            re.IGNORECASE,
        )
        if match is None:
            return None

        r, g, b = (int(v, 16) // 16 ** (len(v) - 2) for v in match.groups())
        foreground = ColorValue.from_rgb(r, g, b)

        response = response[match.end() :]

        # Deal with background color.

        match = re.match(
            r"^\x1b]11;rgb:([0-9a-f]{2,4})/([0-9a-f]{2,4})/([0-9a-f]{2,4})(?:\x1b\\|\a)",
            response,
            re.IGNORECASE,
        )
        if match is None:
            return None

        r, g, b = (int(v, 16) // 16 ** (len(v) - 2) for v in match.groups())
        background = ColorValue.from_rgb(r, g, b)
        luma = (0.2627 * r + 0.6780 * g + 0.0593 * b) / 256

        if luma <= 0.2:
            lightness = Lightness.DARK
        elif luma >= 0.85:
            lightness = Lightness.LIGHT
        else:
            lightness = Lightness.UNKNOWN

        response = response[match.end() :]

        # Deal with other colors

        colors = {}

        while response:
            match = re.match(
                r"^\x1b]4;(\d+);rgb:([0-9a-f]{2,4})/([0-9a-f]{2,4})/([0-9a-f]{2,4})(?:\x1b\\|\a)",
                response,
                re.IGNORECASE,
            )
            if match is None:
                return None

            c = int(match.group(1))
            r, g, b = (int(v, 16) // 16 ** (len(v) - 2) for v in match.groups()[1:])
            colors[c] = ColorValue.from_rgb(r, g, b)

            response = response[match.end() :]

        if set(colors.keys()) != {0, 1, 2, 3, 4, 5, 6, 7}:
            return None

        # return colors
        return TerminalColors(
            background=background,
            foreground=foreground,
            black=colors[0],
            red=colors[1],
            green=colors[2],
            yellow=colors[3],
            blue=colors[4],
            magenta=colors[5],
            cyan=colors[6],
            white=colors[7],
            lightness=lightness,
        )

    except Exception:
        return None


def _query_term(stream: _t.TextIO, query: str) -> str | None:
    try:
        with _enter_raw_mode(stream):
            # Lock the keyboard.
            stream.write("\x1b[2h")
            stream.flush()
            _flush_input_buffer()

            # It is important that we unlock keyboard before exiting `cbreak`,
            # hence the nested `try`.
            try:
                # Append a DA1 query, as virtually all terminals support it.
                stream.write(query + "\x1b[c")
                stream.flush()

                buf = _read_keycode()
                if not buf.startswith("\x1b"):
                    yuio._logger.debug("_query_term invalid response")
                    return None

                # Read till we find a DA1 response.
                while not re.search(r"\x1b\[\?.*?c", buf):
                    buf += _read_keycode()

                return buf[: buf.index("\x1b[?")]
            finally:
                _flush_input_buffer()

                # Release the keyboard.
                stream.write("\x1b[2i")
                stream.flush()
    except Exception:
        yuio._logger.debug("_query_term error", exc_info=True)
        return None


def _is_tty(stream: _t.TextIO | None) -> bool:
    try:
        return stream is not None and stream.isatty()
    except Exception:
        return False


if os.name == "posix":

    def _is_foreground(stream: _t.TextIO | None) -> bool:
        try:
            return stream is not None and os.getpgrp() == os.tcgetpgrp(stream.fileno())
        except Exception:
            return False

elif os.name == "nt":

    def _is_foreground(stream: _t.TextIO | None) -> bool:
        return True

else:

    def _is_foreground(stream: _t.TextIO | None) -> bool:
        return False


def _is_interactive_input(stream: _t.TextIO | None) -> bool:
    try:
        return stream is not None and _is_tty(stream) and stream.readable()
    except Exception:
        return False


def _is_interactive_output(stream: _t.TextIO | None) -> bool:
    try:
        return stream is not None and _is_tty(stream) and stream.writable()
    except Exception:
        return False


# Platform-specific code for working with terminals.
if os.name == "posix":
    import select
    import termios
    import tty

    @contextlib.contextmanager
    def _enter_raw_mode(stream: _t.TextIO, bracketed_paste: bool = False):
        assert sys.__stdin__ is not None

        prev_mode = termios.tcgetattr(sys.__stdin__)
        tty.setcbreak(sys.__stdin__, termios.TCSANOW)

        if bracketed_paste:
            stream.write("\x1b[?2004h")
            stream.flush()

        try:
            yield
        finally:
            if bracketed_paste:
                stream.write("\x1b[?2004l")
                stream.flush()
            termios.tcsetattr(sys.__stdin__, termios.TCSAFLUSH, prev_mode)

    def _read_keycode() -> str:
        assert sys.__stdin__ is not None

        key = os.read(sys.__stdin__.fileno(), 1)
        while bool(select.select([sys.__stdin__], [], [], 0)[0]):
            key += os.read(sys.__stdin__.fileno(), 1)

        return key.decode(sys.__stdin__.encoding, errors="replace")

    def _flush_input_buffer():
        pass

    def _enable_vt_processing(ostream: _t.TextIO) -> bool:
        return False  # This is a windows functionality

elif os.name == "nt":
    import ctypes
    import ctypes.wintypes
    import msvcrt

    _FlushConsoleInputBuffer = ctypes.windll.kernel32.FlushConsoleInputBuffer
    _FlushConsoleInputBuffer.argtypes = [ctypes.wintypes.HANDLE]
    _FlushConsoleInputBuffer.restype = ctypes.wintypes.BOOL

    _GetConsoleMode = ctypes.windll.kernel32.GetConsoleMode
    _GetConsoleMode.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.LPDWORD]
    _GetConsoleMode.restype = ctypes.wintypes.BOOL

    _SetConsoleMode = ctypes.windll.kernel32.SetConsoleMode
    _SetConsoleMode.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.DWORD]
    _SetConsoleMode.restype = ctypes.wintypes.BOOL

    _ReadConsoleW = ctypes.windll.kernel32.ReadConsoleW
    _ReadConsoleW.argtypes = [
        ctypes.wintypes.HANDLE,
        ctypes.wintypes.LPVOID,
        ctypes.wintypes.DWORD,
        ctypes.wintypes.LPDWORD,
        ctypes.wintypes.LPVOID,
    ]
    _ReadConsoleW.restype = ctypes.wintypes.BOOL

    _ENABLE_PROCESSED_OUTPUT = 0x0001
    _ENABLE_WRAP_AT_EOL_OUTPUT = 0x0002
    _ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
    _ENABLE_VIRTUAL_TERMINAL_INPUT = 0x0200

    if sys.__stdin__ is not None:  # TODO: don't rely on sys.__stdin__?
        _ISTREAM_HANDLE = msvcrt.get_osfhandle(sys.__stdin__.fileno())
    else:
        _ISTREAM_HANDLE = None

    @contextlib.contextmanager
    def _enter_raw_mode(stream: _t.TextIO, bracketed_paste: bool = False):
        assert _ISTREAM_HANDLE is not None

        mode = ctypes.wintypes.DWORD()
        success = _GetConsoleMode(_ISTREAM_HANDLE, ctypes.byref(mode))
        if not success:
            raise ctypes.WinError()
        success = _SetConsoleMode(_ISTREAM_HANDLE, _ENABLE_VIRTUAL_TERMINAL_INPUT)
        if not success:
            raise ctypes.WinError()

        try:
            yield
        finally:
            success = _SetConsoleMode(_ISTREAM_HANDLE, mode)
            if not success:
                raise ctypes.WinError()

    def _read_keycode() -> str:
        assert _ISTREAM_HANDLE is not None

        CHAR16 = ctypes.wintypes.WCHAR * 16

        n_read = ctypes.wintypes.DWORD()
        buffer = CHAR16()

        success = _ReadConsoleW(
            _ISTREAM_HANDLE,
            ctypes.byref(buffer),
            16,
            ctypes.byref(n_read),
            0,
        )
        if not success:
            raise ctypes.WinError()

        return buffer.value

    def _flush_input_buffer():
        assert _ISTREAM_HANDLE is not None

        success = _FlushConsoleInputBuffer(_ISTREAM_HANDLE)
        if not success:
            raise ctypes.WinError()

    def _enable_vt_processing(ostream: _t.TextIO) -> bool:
        try:
            version = sys.getwindowsversion()
            if version.major < 10 or version.build < 14931:
                return False

            handle = msvcrt.get_osfhandle(ostream.fileno())
            if not bool(
                _SetConsoleMode(
                    handle,
                    _ENABLE_PROCESSED_OUTPUT
                    | _ENABLE_WRAP_AT_EOL_OUTPUT
                    | _ENABLE_VIRTUAL_TERMINAL_PROCESSING,
                )
            ):
                return False

            return True
        except Exception:
            return False

else:

    @contextlib.contextmanager
    def _enter_raw_mode(stream: _t.TextIO, bracketed_paste: bool = False):
        raise OSError("not supported")
        yield

    def _read_keycode() -> str:
        raise OSError("not supported")

    def _flush_input_buffer():
        raise OSError("not supported")

    def _enable_vt_processing(ostream: _t.TextIO) -> bool:
        raise OSError("not supported")


@dataclass(frozen=True, **yuio._with_slots())
class ColorValue:
    """
    Data about a single color.

    Can be either a single ANSI escape code, or an RGB-tuple.

    Single ANSI escape code represents a standard terminal color code.
    The actual color value for it is controlled by the terminal's user.
    Therefore, it doesn't permit operations on colors,
    such as :meth:`ColorValue.darken` or :meth:`ColorValue.lerp`.

    An RGB-tuple represents a true color. When displaying on a terminal that
    doesn't support true colors, it will be converted to a corresponding
    256-color or an 8-color automatically.

    """

    data: int | str | tuple[int, int, int]
    """
    Color data.

    """

    @classmethod
    def from_rgb(cls, r: int, g: int, b: int, /) -> ColorValue:
        """
        Create a color value from rgb components.

        Each component should be between 0 and 255.

        Example::

            >>> ColorValue.from_rgb(0xA0, 0x1E, 0x9C)
            <ColorValue #A01E9C>

        """

        return cls((r, g, b))

    @classmethod
    def from_hex(cls, h: str, /) -> ColorValue:
        """
        Create a color value from a hex string.

        Example::

            >>> ColorValue.from_hex('#A01E9C')
            <ColorValue #A01E9C>

        """

        return cls(_parse_hex(h))

    def to_hex(self) -> str | None:
        """
        Return color in hex format with leading ``#``.

        Example::

            >>> a = ColorValue.from_hex('#A01E9C')
            >>> a.to_hex()
            '#A01E9C'

        """

        rgb = self.to_rgb()
        if rgb is not None:
            return f"#{rgb[0]:02X}{rgb[1]:02X}{rgb[2]:02X}"
        else:
            return None

    def to_rgb(self) -> tuple[int, int, int] | None:
        """
        Return RGB components of the color.

        Example::

            >>> a = ColorValue.from_hex('#A01E9C')
            >>> a.to_rgb()
            (160, 30, 156)
        """

        if isinstance(self.data, int):
            return _8_to_rgb(self.data)
        elif isinstance(self.data, tuple):
            return self.data
        else:
            return None

    def darken(self, amount: float, /) -> ColorValue:
        """
        Make this color darker by the given percentage.

        Amount should be between 0 and 1.

        Example::

            >>> # Darken by 30%.
            ... ColorValue.from_hex('#A01E9C').darken(0.30)
            <ColorValue #70156D>

        """

        rgb = self.to_rgb()
        if rgb is None:
            return self

        amount = max(min(amount, 1), 0)
        r, g, b = rgb
        h, s, v = colorsys.rgb_to_hsv(r / 0xFF, g / 0xFF, b / 0xFF)
        v = v - v * amount
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        return ColorValue.from_rgb(int(r * 0xFF), int(g * 0xFF), int(b * 0xFF))

    def lighten(self, amount: float, /) -> ColorValue:
        """
        Make this color lighter by the given percentage.

        Amount should be between 0 and 1.

        Example::

            >>> # Lighten by 30%.
            ... ColorValue.from_hex('#A01E9C').lighten(0.30)
            <ColorValue #BC23B7>

        """

        rgb = self.to_rgb()
        if rgb is None:
            return self

        amount = max(min(amount, 1), 0)
        r, g, b = rgb
        h, s, v = colorsys.rgb_to_hsv(r / 0xFF, g / 0xFF, b / 0xFF)
        v = 1 - v
        v = 1 - (v - v * amount)
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        return ColorValue.from_rgb(int(r * 0xFF), int(g * 0xFF), int(b * 0xFF))

    def match_luminosity(self, other: ColorValue, /) -> ColorValue:
        """
        Set luminosity of this color equal to one of the other color.

        This function will keep hue and saturation of the color intact,
        but it will become as bright as the other color.

        """

        rgb1, rgb2 = self.to_rgb(), other.to_rgb()
        if rgb1 is None or rgb2 is None:
            return self

        h, s, _ = colorsys.rgb_to_hsv(rgb1[0] / 0xFF, rgb1[1] / 0xFF, rgb1[2] / 0xFF)
        _, _, v = colorsys.rgb_to_hsv(rgb2[0] / 0xFF, rgb2[1] / 0xFF, rgb2[2] / 0xFF)
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        return ColorValue.from_rgb(int(r * 0xFF), int(g * 0xFF), int(b * 0xFF))

    @staticmethod
    def lerp(*colors: ColorValue) -> _t.Callable[[float], ColorValue]:
        """
        Return a lambda that allows linear interpolation between several colors.

        If either color is a single ANSI escape code, the first color is always returned
        from the lambda.

        Example::

            >>> a = ColorValue.from_hex('#A01E9C')
            >>> b = ColorValue.from_hex('#22C60C')
            >>> lerp = ColorValue.lerp(a, b)

            >>> lerp(0)
            <ColorValue #A01E9C>
            >>> lerp(0.5)
            <ColorValue #617254>
            >>> lerp(1)
            <ColorValue #22C60C>

        """

        if not colors:
            raise TypeError("lerp expected at least 1 argument, got 0")
        elif len(colors) == 1 or not all(
            isinstance(color.data, tuple) for color in colors
        ):
            return lambda f, /: colors[0]
        else:
            l = len(colors) - 1

            def lerp(f: float, /) -> ColorValue:
                i = int(f * l)
                f = (f - (i / l)) * l

                if i == l:
                    return colors[l]
                else:
                    a, b = colors[i].data, colors[i + 1].data
                    return ColorValue(tuple(int(ca + f * (cb - ca)) for ca, cb in zip(a, b)))  # type: ignore

            return lerp

    def _as_fore(self, term: Term, /) -> str:
        return self._as_code(term, fg_bg_prefix="3")

    def _as_back(self, term: Term, /) -> str:
        return self._as_code(term, fg_bg_prefix="4")

    def _as_code(self, term: Term, /, fg_bg_prefix: str) -> str:
        if not term.has_colors:
            return ""
        elif isinstance(self.data, int):
            return f"{fg_bg_prefix}{self.data}"
        elif isinstance(self.data, str):
            return self.data
        elif term.has_colors_true:
            return f"{fg_bg_prefix}8;2;{self.data[0]};{self.data[1]};{self.data[2]}"
        elif term.has_colors_256:
            return f"{fg_bg_prefix}8;5;{_rgb_to_256(*self.data)}"
        else:
            return f"{fg_bg_prefix}{_rgb_to_8(*self.data)}"

    def __repr__(self) -> str:
        if isinstance(self.data, tuple):
            return f"<ColorValue {self.to_hex()}>"
        else:
            return f"<ColorValue {self.data}>"


@dataclass(frozen=True)
class Color:
    """
    Data about terminal output style. Contains
    foreground and background color, as well as text styles.

    This class only contains data about the color. It doesn't know anything about
    how to apply it to a terminal, or whether a terminal supports colors at all.
    These decisions are thus deferred to a lower level of API (see :meth:`Color.as_code`).

    When converted to an ANSI code and printed, a color completely overwrites a previous
    color that was used by a terminal. This behavior prevents different colors and styles
    bleeding one into another. So, for example, printing :data:`Color.STYLE_BOLD`
    and then :data:`Color.FORE_RED` will result in non-bold red text.

    Colors can be combined before printing, though::

        >>> Color.STYLE_BOLD | Color.FORE_RED  # Bold red
        Color(fore=<ColorValue 1>, back=None, bold=True, dim=None)

    Yuio supports true RGB colors. They are automatically converted
    to 256- or 8-bit colors if needed.

    """

    fore: ColorValue | None = None
    """
    Foreground color.

    """

    back: ColorValue | None = None
    """
    Background color.

    """

    bold: bool | None = None
    """
    If true, render text as bold.

    """

    dim: bool | None = None
    """
    If true, render text as dim.

    """

    def __or__(self, other: Color, /):
        return Color(
            other.fore if other.fore is not None else self.fore,
            other.back if other.back is not None else self.back,
            other.bold if other.bold is not None else self.bold,
            other.dim if other.dim is not None else self.dim,
        )

    def __ior__(self, other: Color, /):
        return self | other

    @classmethod
    def fore_from_rgb(cls, r: int, g: int, b: int) -> Color:
        """
        Create a foreground color value from rgb components.

        Each component should be between 0 and 255.

        Example::

            >>> Color.fore_from_rgb(0xA0, 0x1E, 0x9C)
            Color(fore=<ColorValue #A01E9C>, back=None, bold=None, dim=None)

        """

        return cls(fore=ColorValue.from_rgb(r, g, b))

    @classmethod
    def fore_from_hex(cls, h: str) -> Color:
        """
        Create a foreground color value from a hex string.

        Example::

            >>> Color.fore_from_hex('#A01E9C')
            Color(fore=<ColorValue #A01E9C>, back=None, bold=None, dim=None)

        """

        return cls(fore=ColorValue.from_hex(h))

    @classmethod
    def back_from_rgb(cls, r: int, g: int, b: int) -> Color:
        """
        Create a background color value from rgb components.

        Each component should be between 0 and 255.

        Example::

            >>> Color.back_from_rgb(0xA0, 0x1E, 0x9C)
            Color(fore=None, back=<ColorValue #A01E9C>, bold=None, dim=None)

        """

        return cls(back=ColorValue.from_rgb(r, g, b))

    @classmethod
    def back_from_hex(cls, h: str) -> Color:
        """
        Create a background color value from a hex string.

        Example::

            >>> Color.back_from_hex('#A01E9C')
            Color(fore=None, back=<ColorValue #A01E9C>, bold=None, dim=None)

        """

        return cls(back=ColorValue.from_hex(h))

    @staticmethod
    def lerp(*colors: Color) -> _t.Callable[[float], Color]:
        """
        Return a lambda that allows linear interpolation between several colors.

        If either color is a single ANSI escape code, the first color is always returned
        from the lambda.

        Example::

            >>> a = Color.fore_from_hex('#A01E9C')
            >>> b = Color.fore_from_hex('#22C60C')
            >>> lerp = Color.lerp(a, b)

            >>> lerp(0)
            Color(fore=<ColorValue #A01E9C>, back=None, bold=None, dim=None)
            >>> lerp(0.5)
            Color(fore=<ColorValue #617254>, back=None, bold=None, dim=None)
            >>> lerp(1)
            Color(fore=<ColorValue #22C60C>, back=None, bold=None, dim=None)

        """

        if not colors:
            raise TypeError("lerp expected at least 1 argument, got 0")
        elif len(colors) == 1:
            return lambda f, /: colors[0]
        else:
            fore_lerp = all(
                color.fore is not None and isinstance(color.fore.data, tuple)
                for color in colors
            )
            if fore_lerp:
                fore = ColorValue.lerp(*(color.fore for color in colors))  # type: ignore

            back_lerp = all(
                color.back is not None and isinstance(color.back.data, tuple)
                for color in colors
            )
            if back_lerp:
                back = ColorValue.lerp(*(color.back for color in colors))  # type: ignore

            if fore_lerp and back_lerp:
                return lambda f: dataclasses.replace(colors[0], fore=fore(f), back=back(f))  # type: ignore
            elif fore_lerp:
                return lambda f: dataclasses.replace(colors[0], fore=fore(f))  # type: ignore
            elif back_lerp:
                return lambda f: dataclasses.replace(colors[0], back=back(f))  # type: ignore
            else:
                return lambda f, /: colors[0]

    def as_code(self, term: Term, /) -> str:
        """
        Convert this color into an ANSI escape code
        with respect to the given terminal capabilities.

        """

        if not term.has_colors:
            return ""

        codes = []
        if self.fore:
            codes.append(self.fore._as_fore(term))
        if self.back:
            codes.append(self.back._as_back(term))
        if self.bold:
            codes.append("1")
        if self.dim:
            codes.append("2")
        if codes:
            return "\x1b[;" + ";".join(codes) + "m"
        else:
            return "\x1b[m"

    NONE: typing.ClassVar[Color] = dict()  # type: ignore
    """
    No color.

    """

    STYLE_BOLD: typing.ClassVar[Color] = dict(bold=True)  # type: ignore
    """
    Bold font style.

    """

    STYLE_DIM: typing.ClassVar[Color] = dict(dim=True)  # type: ignore
    """
    Dim font style.

    """

    STYLE_NORMAL: typing.ClassVar[Color] = dict(bold=False, dim=False)  # type: ignore
    """
    Not bold nor dim.

    """

    FORE_NORMAL: typing.ClassVar[Color] = dict(fore=ColorValue(9))  # type: ignore
    """
    Normal foreground color.

    """

    FORE_NORMAL_DIM: typing.ClassVar[Color] = dict(fore=ColorValue("2"))  # type: ignore
    """
    Normal foreground color rendered with dim setting.

    This is an alternative to bright black that works with
    most terminals and color schemes.

    """

    FORE_BLACK: typing.ClassVar[Color] = dict(fore=ColorValue(0))  # type: ignore
    """
    Black foreground color.

    .. warning::

       Avoid using this color, in most terminals it is the same as background color.
       Instead, use :attr:`~Color.FORE_NORMAL_DIM`.

    """

    FORE_RED: typing.ClassVar[Color] = dict(fore=ColorValue(1))  # type: ignore
    """
    Red foreground color.

    """

    FORE_GREEN: typing.ClassVar[Color] = dict(fore=ColorValue(2))  # type: ignore
    """
    Green foreground color.

    """

    FORE_YELLOW: typing.ClassVar[Color] = dict(fore=ColorValue(3))  # type: ignore
    """
    Yellow foreground color.

    """

    FORE_BLUE: typing.ClassVar[Color] = dict(fore=ColorValue(4))  # type: ignore
    """
    Blue foreground color.

    """

    FORE_MAGENTA: typing.ClassVar[Color] = dict(fore=ColorValue(5))  # type: ignore
    """
    Magenta foreground color.

    """

    FORE_CYAN: typing.ClassVar[Color] = dict(fore=ColorValue(6))  # type: ignore
    """
    Cyan foreground color.

    """

    FORE_WHITE: typing.ClassVar[Color] = dict(fore=ColorValue(7))  # type: ignore
    """
    White foreground color.

    Avoid using it, in some terminals, notably in the Mac OS default terminal,
    it is unreadable.

    """

    BACK_NORMAL: typing.ClassVar[Color] = dict(back=ColorValue(9))  # type: ignore
    """
    Normal background color.

    """

    BACK_BLACK: typing.ClassVar[Color] = dict(back=ColorValue(0))  # type: ignore
    """
    Black background color.

    """

    BACK_RED: typing.ClassVar[Color] = dict(back=ColorValue(1))  # type: ignore
    """
    Red background color.

    """

    BACK_GREEN: typing.ClassVar[Color] = dict(back=ColorValue(2))  # type: ignore
    """
    Green background color.

    """

    BACK_YELLOW: typing.ClassVar[Color] = dict(back=ColorValue(3))  # type: ignore
    """
    Yellow background color.

    """

    BACK_BLUE: typing.ClassVar[Color] = dict(back=ColorValue(4))  # type: ignore
    """
    Blue background color.

    """

    BACK_MAGENTA: typing.ClassVar[Color] = dict(back=ColorValue(5))  # type: ignore
    """
    Magenta background color.

    """

    BACK_CYAN: typing.ClassVar[Color] = dict(back=ColorValue(6))  # type: ignore
    """
    Cyan background color.

    """

    BACK_WHITE: typing.ClassVar[Color] = dict(back=ColorValue(7))  # type: ignore
    """
    White background color.

    """


for _n, _v in vars(Color).items():
    if _n == _n.upper():
        setattr(Color, _n, Color(**_v))
del _n, _v  # type: ignore


def _rgb_to_256(r: int, g: int, b: int) -> int:
    closest_idx = lambda x, vals: min((abs(x - v), i) for i, v in enumerate(vals))[1]
    color_components = [0x00, 0x5F, 0x87, 0xAF, 0xD7, 0xFF]

    if r == g == b:
        i = closest_idx(r, color_components + [0x08 + 10 * i for i in range(24)])
        if i >= len(color_components):
            return 232 + i - len(color_components)
        r, g, b = i, i, i
    else:
        r, g, b = (closest_idx(x, color_components) for x in (r, g, b))
    return r * 36 + g * 6 + b + 16


def _rgb_to_8(r: int, g: int, b: int) -> int:
    return (
        (1 if r >= 128 else 0)
        | (1 if g >= 128 else 0) << 1
        | (1 if b >= 128 else 0) << 2
    )


def _8_to_rgb(code: int) -> tuple[int, int, int]:
    return (
        (code & 1) and 0xFF,
        (code & 2) and 0xFF,
        (code & 4) and 0xFF,
    )


def _parse_hex(h: str) -> tuple[int, int, int]:
    if not re.match(r"^#[0-9a-fA-F]{6}$", h):
        raise ValueError(f"invalid hex string {h!r}")
    return tuple(int(h[i : i + 2], 16) for i in (1, 3, 5))  # type: ignore


def line_width(s: str, /) -> int:
    """
    Calculates string width when the string is displayed
    in a terminal.

    This function makes effort to detect wide characters
    such as emojis. If does not, however, work correctly
    with extended grapheme clusters, and so it may fail
    for emojis with modifiers, or other complex characters.

    Example where it fails is ``👩🏽‍💻``. It consists
    of four code points:

    - Unicode Character `WOMAN` (``U+1F469``, ``👩``),
    - Unicode Character `EMOJI MODIFIER FITZPATRICK TYPE-4` (``U+1F3FD``),
    - Unicode Character `ZERO WIDTH JOINER` (``U+200D``),
    - Unicode Character `PERSONAL COMPUTER` (``U+1F4BB``, ``💻``).

    Since :func:`line_width` can't understand that these code points
    are combined into a single emoji, it treats them separately,
    resulting in answer `6` (`2` for every code point except `ZERO WIDTH JOINER`)::

        >>> line_width("\U0001f469\U0001f3fd\U0000200d\U0001f4bb")
        6

    In all fairness, detecting how much space such an emoji will take
    is not so straight forward, as that will depend on unicode capabilities
    of a specific terminal. Since a lot of terminals will not handle such emojis
    correctly, I've decided to go with this simplistic implementation.

    """

    # Note: it may be better to bundle `wcwidth` and use it instead of the code below.
    # However, there is an issue that `wcwidth`'s results are not additive.
    # In the above example, `wcswidth('👩🏽‍💻')` will see that it is two-spaces wide,
    # while `sum(wcwidth(c) for c in '👩🏽‍💻')` will report that it is four-spaces wide.
    # To render it properly, the widget will have to be aware of extended grapheme
    # clusters, and generally this will be a lot of headache. Since most terminals
    # won't handle these edge cases correctly, I don't want to bother.

    if s.isascii():
        # Fast path. Note that our renderer replaces unprintable characters
        # with spaces, so ascii strings always have width equal to their length.
        return len(s)
    else:
        # Long path. It kinda works, but not always, but most of the times...
        return sum(
            (unicodedata.east_asian_width(c) in "WF") + 1
            for c in s
            if unicodedata.category(c)[0] not in "MC"
        )


class NoWrap(str):
    """
    A string that will not be wrapped by the text wrapper.

    See :meth:`ColorizedString.wrap` for more info.

    """


class Esc(NoWrap):
    """
    A string that signifies an escaped special symbol.

    This string is not wrapped even if ``break_long_nowrap_words`` is :data:`True`.

    """


@_t.final
class ColorizedString:
    """
    A string with colors.

    This class is a wrapper over a list of strings and colors.
    Each color applies to strings after it, right until the next color.

    :class:`ColorizedString` supports some basic string operations.
    Most notable, it supports wide-character-aware wrapping (see :func:`line_width`),
    and ``%``-formatting.

    Unlike :class:`str` instances, :class:`ColorizedString` is mutable through
    the ``+=`` operator.

    You can build a colorized string from raw parts,
    or you can use :meth:`yuio.md.colorize`.

    """

    def __init__(
        self,
        content: AnyString = "",
        /,
        *,
        explicit_newline: str = "",
    ):
        self._parts: list[Color | str]
        if isinstance(content, (str, Color)):
            self._parts = [content] if content else []
        elif isinstance(content, ColorizedString):
            self._parts = list(content._parts)
        else:
            self._parts = list(content)

        self._explicit_newline = explicit_newline

    @property
    def explicit_newline(self) -> str:
        """
        Explicit newline indicates that a line of a wrapped text
        was broken because the original text contained a new line character.

        See :meth:`~ColorizedString.wrap` for details.

        """

        return self._explicit_newline

    @functools.cached_property
    def width(self) -> int:
        """
        String width when the string is displayed in a terminal.

        See :func:`line_width` for more information.

        """

        return sum(line_width(s) for s in self._parts if isinstance(s, str))

    @functools.cached_property
    def len(self) -> int:
        """
        Line length in bytes, ignoring all colors.

        """

        return sum(len(s) for s in self._parts if isinstance(s, str))

    def __len__(self) -> int:
        return self.len

    def __bool__(self) -> bool:
        return self.len > 0

    def iter(self) -> _t.Iterator[Color | str]:
        """
        Iterate over raw parts of the string,
        i.e. the underlying list of strings and colors.

        """

        return self._parts.__iter__()

    def __iter__(self) -> _t.Iterator[Color | str]:
        return self.iter()

    def wrap(
        self,
        width: int,
        /,
        *,
        preserve_spaces: bool = False,
        preserve_newlines: bool = True,
        break_long_words: bool = True,
        break_long_nowrap_words: bool = False,
        first_line_indent: AnyString = "",
        continuation_indent: AnyString = "",
    ) -> list[ColorizedString]:
        r"""
        Wrap a long line of text into multiple lines.

        :param preserve_spaces:
            if set to :data:`True`, all spaces are preserved.
            Otherwise, consecutive spaces are collapsed into a single space.
            Note that tabs are always treated as a single space.
        :param preserve_newlines:
            if set to :data:`True` (default), text is additionally wrapped
            on newline characters. When this happens, the newline sequence that wrapped
            the line will be placed into :attr:`~ColorizedString.explicit_newline`.

            If set to :data:`False`, newlines are treated as whitespaces.
        :param break_long_words:
            if set to :data:`True` (default), words that don't fit into a single line
            will be split into multiple lines.
        :param break_long_nowrap_words:
            if set to :data:`True`, :class:`NoWrap` words that don't fit
            into a single line will be split into multiple lines.
        :param first_line_indent:
            a string that will be prepended before the first line.
        :param continuation_indent:
            a string that will be prepended before all subsequent lines.

        Example::

            >>> ColorizedString("hello, world!\nit's a good day!").wrap(13)  # doctest: +NORMALIZE_WHITESPACE
            [<ColorizedString('hello, world!', explicit_newline='\n')>,
             <ColorizedString("it's a good")>,
             <ColorizedString('day!')>]

        """

        return _TextWrapper(
            width,
            preserve_spaces=preserve_spaces,
            preserve_newlines=preserve_newlines,
            break_long_words=break_long_words,
            break_long_nowrap_words=break_long_nowrap_words,
            first_line_indent=first_line_indent,
            continuation_indent=continuation_indent,
        ).wrap(self)

    def indent(
        self,
        first_line_indent: AnyString = "",
        continuation_indent: AnyString = "",
    ) -> ColorizedString:
        r"""
        Indent this string by the given sequence.

        :param first_line_indent:
            this will be appended to the first line in the string.
        :param continuation_indent:
            this will be appended to subsequent lines in the string.

        Example::

            >>> ColorizedString("hello, world!\nit's a good day!").indent("# ", "  ")
            <ColorizedString("# hello, world!\n  it's a good day!")>

        """

        res = ColorizedString()

        color: Color = Color.NONE
        cur_color: Color = Color.NONE
        needs_indent = True

        for part in self._parts:
            if isinstance(part, Color):
                color = part
                continue

            for line in part.splitlines(keepends=True):
                if needs_indent:
                    if cur_color != Color.NONE:
                        res += Color.NONE
                    res += first_line_indent
                    first_line_indent = ColorizedString(continuation_indent)

                if color and (needs_indent or color != cur_color):
                    res += color
                    cur_color = color

                res += line

                needs_indent = line.endswith("\n")

        if color and color != cur_color:
            res += color

        return res

    def percent_format(self, args: _t.Any) -> ColorizedString:
        """
        Format colorized string as if with ``%``-formatting
        (i.e. `old-style formatting`_).

        .. _old-style formatting: https://docs.python.org/3/library/stdtypes.html#printf-style-string-formatting

        Example::

            >>> line = ColorizedString("Hello, %s!")
            >>> line % "Username"
            <ColorizedString('Hello, Username!')>

        """

        return ColorizedString(_percent_format(self, args))

    def __mod__(self, args: _t.Any) -> ColorizedString:
        return self.percent_format(args)

    def __imod__(self, args: _t.Any) -> ColorizedString:
        self._parts = _percent_format(self, args)

        self.__dict__.pop("width", None)
        self.__dict__.pop("len", None)

        return self

    def __add__(self, rhs: AnyString) -> ColorizedString:
        copy = ColorizedString(self)
        copy += rhs
        return copy

    def __radd__(self, lhs: AnyString) -> ColorizedString:
        copy = ColorizedString(lhs)
        copy += self
        return copy

    def __iadd__(self, rhs: AnyString) -> ColorizedString:
        if isinstance(rhs, (str, Color)):
            if rhs:
                self._parts.append(rhs)
        elif isinstance(rhs, ColorizedString):
            self._parts.extend(rhs._parts)
        else:
            self._parts.extend(rhs)

        self.__dict__.pop("width", None)
        self.__dict__.pop("len", None)

        return self

    def process_colors(self, term: Term, /) -> list[str]:
        """
        Convert colors in this string into a normal string
        with ANSI escape sequences.

        """

        out: list[str] = []

        color: Color | None = None
        cur_color: Color | None = None

        for part in self._parts:
            if isinstance(part, Color):
                color = part
            else:
                if color and color != cur_color:
                    out.append(color.as_code(term))
                    cur_color = color
                out.append(part)

        if color and color != cur_color:
            out.append(color.as_code(term))

        return out

    def __str__(self) -> str:
        return "".join(c for c in self._parts if isinstance(c, str))

    def __repr__(self) -> str:
        if self.explicit_newline:
            return f"<ColorizedString({str(self)!r}, explicit_newline={self.explicit_newline!r})>"
        else:
            return f"<ColorizedString({str(self)!r})>"


if _t.TYPE_CHECKING:

    RawString: _t.TypeAlias = _t.Iterable[Color | str]
    """
    Raw colorized string. This is the underlying type for :class:`ColorizedString`.

    """

    AnyString: _t.TypeAlias = _t.Union[str, Color, RawString, ColorizedString]  # type: ignore
    """
    Any string (i.e. a :class:`str`, a raw colorized string, or a normal colorized string).

    """

elif "__YUIO_SPHINX_BUILD" in os.environ:

    class RawString: ...

    class AnyString: ...

else:

    RawString: _t.TypeAlias = lambda x: x  # type: ignore

    AnyString: _t.TypeAlias = lambda x: x  # type: ignore


_S_SYNTAX = re.compile(
    r"""
        %                               # Percent
        (?:\((?P<mapping>[^)]*)\))?     # Mapping key
        (?P<flag>[#0\-+ ])*             # Conversion Flag
        (?P<width>\*|\d+)?              # Field width
        (?:\.(?P<precision>\*|\d*))?    # Precision
        [hlL]?                          # Unused length modifier
        (?P<format>.)                   # Conversion type
    """,
    re.VERBOSE,
)


def _percent_format(s: ColorizedString, args: _t.Any) -> list[Color | str]:
    if not isinstance(args, (dict, tuple)):
        args = (args,)

    i = 0

    def repl(m: _t.StrReMatch) -> str:
        nonlocal i
        groups = m.groupdict()
        if groups["format"] == "%":
            if m.group(0) != "%%":
                raise ValueError("unsupported format character '%'")
            return "%"

        if groups["mapping"] is not None:
            fmt_args = args
        elif isinstance(args, tuple):
            begin = i
            end = i = i + 1 + (m.group("width") == "*") + (m.group("precision") == "*")
            fmt_args = args[begin:end]
        elif i == 0:
            # We've passed a dict, and now want to format it with `%s`.
            # We allow that once. I.e. `"%s" % {}` is fine, `"%s %s" % {}` is not.
            fmt_args = args
            i = 1
        else:
            raise TypeError("not enough arguments for format string")

        return m.group(0) % fmt_args

    raw = []
    for part in s:
        if isinstance(part, NoWrap):
            raw.append(NoWrap(_S_SYNTAX.sub(repl, part)))
        elif isinstance(part, str):
            raw.append(_S_SYNTAX.sub(repl, part))
        else:
            raw.append(part)

    if isinstance(args, tuple) and i < len(args):
        raise TypeError("not all arguments converted during string formatting")

    return raw


_SPACE_TRANS = str.maketrans("\r\n\t\v\b\f", "      ")

_WORD_PUNCT = r'[\w!"\'&.,?]'
_LETTER = r"[^\d\W]"
_NOWHITESPACE = r"[^ \r\n\t\v\b\f]"

# Copied from textwrap with some modifications in newline handling
_WORDSEP_RE = re.compile(
    r"""
    ( # newlines and line feeds are matched one-by-one
        \v?(?:\r\n|\r|\n)
    | # any whitespace
        [ \t\v\b\f]+
    | # em-dash between words
        (?<=%(wp)s) -{2,} (?=\w)
    | # word, possibly hyphenated
        %(nws)s+? (?:
        # hyphenated word
            -(?: (?<=%(lt)s{2}-) | (?<=%(lt)s-%(lt)s-))
            (?= %(lt)s -? %(lt)s)
        | # end of word
            (?=[ \r\n\t\v\b\f]|\Z)
        | # em-dash
            (?<=%(wp)s) (?=-{2,}\w)
        )
    )"""
    % {"wp": _WORD_PUNCT, "lt": _LETTER, "nws": _NOWHITESPACE},
    re.VERBOSE,
)
_WORDSEP_NL_RE = re.compile(r"(\v?(?:\r\n|\r|\n))")


class _TextWrapper:
    def __init__(
        self,
        width: int,
        /,
        *,
        preserve_spaces: bool,
        preserve_newlines: bool,
        break_long_words: bool,
        break_long_nowrap_words: bool,
        first_line_indent: AnyString = "",
        continuation_indent: AnyString = "",
    ):
        self.width: int = width
        self.preserve_spaces: bool = preserve_spaces
        self.preserve_newlines: bool = preserve_newlines
        self.break_long_words: bool = break_long_words
        self.break_long_nowrap_words: bool = break_long_nowrap_words
        self.first_line_indent: ColorizedString = ColorizedString(first_line_indent)
        self.continuation_indent: ColorizedString = ColorizedString(continuation_indent)

        if (
            self.width - self.first_line_indent.width <= 1
            or self.width - self.continuation_indent.width <= 1
        ):
            self.width = (
                max(self.first_line_indent.width, self.continuation_indent.width) + 2
            )

        self.lines: list[ColorizedString] = []

        self.current_line: list[Color | str] = [Color.NONE]
        if self.first_line_indent:
            self.current_line.extend(list(self.first_line_indent))
            self.current_line.append(Color.NONE)
        self.current_line_width: int = self.first_line_indent.width
        self.current_color: Color = Color.NONE
        self.current_line_is_nonempty: bool = False

    def _flush_line(self, explicit_newline=""):
        self.lines.append(
            ColorizedString(self.current_line, explicit_newline=explicit_newline)
        )
        self.current_line: list[Color | str] = []
        if self.continuation_indent:
            self.current_line.append(Color.NONE)
            self.current_line.extend(list(self.continuation_indent))
        self.current_line.append(self.current_color)
        self.current_line_width: int = self.continuation_indent.width
        self.current_line_is_nonempty = False

    def _append_word(self, word: str, word_width: int):
        self.current_line_is_nonempty = True
        self.current_line.append(word)
        self.current_line_width += word_width

    def _append_color(self, color: Color):
        if color != self.current_color:
            self.current_color = color
            self.current_line.append(color)

    def _append_word_with_breaks(self, word: str, word_width: int):
        while self.current_line_width + word_width > self.width:
            word_head_len = word_head_width = 0

            for c in word:
                c_width = line_width(c)
                if self.current_line_width + word_head_width + c_width > self.width:
                    break
                word_head_len += 1
                word_head_width += c_width

            self._append_word(word[:word_head_len], word_head_width)

            word = word[word_head_len:]
            word_width -= word_head_width

            self._flush_line()

        if word:
            self._append_word(word, word_width)

    def wrap(self, text: ColorizedString) -> list[ColorizedString]:
        need_space_before_word = False
        at_line_beginning = True

        for part in text:
            if isinstance(part, Color):
                if (
                    need_space_before_word
                    and self.current_line_width + need_space_before_word <= self.width
                ):
                    # Make sure any space that was issued before the color is flushed.
                    self._append_word(" ", 1)
                    need_space_before_word = False
                self._append_color(part)
                continue

            nowrap = False
            if isinstance(part, Esc):
                words = [part]
            elif isinstance(part, NoWrap):
                words = _WORDSEP_NL_RE.split(part)
                nowrap = True
            else:
                words = _WORDSEP_RE.split(part)

            for word in words:
                if not word:
                    continue

                if "\v" in word or (
                    word in ("\r", "\n", "\r\n") and self.preserve_newlines
                ):
                    self._flush_line(explicit_newline=word)
                    need_space_before_word = False
                    at_line_beginning = True
                    continue

                if word.isspace():
                    if nowrap:
                        word = word.translate(_SPACE_TRANS)
                        self._append_word(word, len(word))
                    elif at_line_beginning or self.preserve_spaces:
                        word = word.translate(_SPACE_TRANS)
                        self._append_word_with_breaks(word, len(word))
                    else:
                        need_space_before_word = True
                    continue

                word_width = line_width(word)

                if (
                    self.current_line_width + word_width + need_space_before_word
                    <= self.width
                ):
                    # Word fits onto the current line.
                    if need_space_before_word:
                        self._append_word(" ", 1)
                    self._append_word(word, word_width)
                else:
                    # Word doesn't fit, so we start a new line.
                    if self.current_line_is_nonempty:
                        self._flush_line()
                    if (
                        (nowrap and self.break_long_nowrap_words)
                        or (not nowrap and self.break_long_words)
                        and not isinstance(word, Esc)
                    ):
                        # We will break the word in the middle if it doesn't fit
                        # onto the whole line.
                        self._append_word_with_breaks(word, word_width)
                    else:
                        self._append_word(word, word_width)

                need_space_before_word = False
                at_line_beginning = False

        if self.current_line or not self.lines or self.lines[-1].explicit_newline:
            self._flush_line()

        return self.lines



import yuio.app
import yuio.io
from time import sleep

@yuio.app.app
def main():
    with yuio.io.Task("Sending a greeting"):
        send_greeting()

def send_greeting():
    sleep(5)  # Sending greeting requires five seconds.

if __name__ == "__main__":
    main.run()



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides a base class for configs that can be loaded from
files, environment variables or command line arguments (via :mod:`yuio.app`).

Derive your config from the :class:`Config` base class. Inside of its body,
define config fields using type annotations, just like :mod:`dataclasses`:

.. code-block:: python

    class AppConfig(Config):
        #: trained model to execute
        model: pathlib.Path

        #: input data for the model
        data: pathlib.Path

        #: enable or disable gpu
        use_gpu: bool = True

Then use config's constructors and the :meth:`~Config.update` method
to load it from various sources::

    # Load config from a file.
    config = AppConfig.load_from_json_file("~/.my_app_cfg.json")

    # Update config with values from env.
    config.update(AppConfig.load_from_env())


Config base class
-----------------

.. autoclass:: Config


Advanced field configuration
----------------------------

By default, :class:`Config` infers names for env variables and flags,
appropriate parsers, and other things from field's name, type hint, and comments.

If you need to override them, you can use :func:`yuio.app.field`
and :func:`yuio.app.inline` functions (also available from ``yuio.config.field``
and ``yuio.config.inline``):

.. code-block:: python

    class AppConfig(Config):
        model: pathlib.Path | None = field(
            default=None,
            help="trained model to execute",
        )


Nesting configs
---------------

You can nest configs to achieve modularity:

.. code-block:: python

    class ExecutorConfig(Config):
        #: number of threads to use
        threads: int

        #: enable or disable gpu
        use_gpu: bool = True

    class AppConfig(Config):
        #: executor parameters
        executor: ExecutorConfig

        #: trained model to execute
        model: pathlib.Path

To initialise a nested config, pass either an instance of if
or a dict with its variables to the config's constructor:

.. code-block:: python

    # The following lines are equivalent:
    config = AppConfig(executor=ExecutorConfig(threads=16))
    config = AppConfig(executor={'threads': 16})
    # ...although type checkers will complain about dict =(


Parsing environment variables
-----------------------------

You can load config from environment through :meth:`~Config.load_from_env`.

Names of environment variables are just capitalized field names.
Use the :func:`yuio.app.field` function to override them:

.. code-block:: python

    class KillCmdConfig(Config):
        # Will be loaded from `SIGNAL`.
        signal: int

        # Will be loaded from `PROCESS_ID`.
        pid: int = field(env='PROCESS_ID')

In nested configs, environment variable names are prefixed with name
of a field that contains the nested config:

.. code-block:: python

    class BigConfig(Config):
        # `kill_cmd.signal` will be loaded from `KILL_CMD_SIGNAL`.
        kill_cmd: KillCmdConfig

        # `kill_cmd_2.signal` will be loaded from `KILL_SIGNAL`.
        kill_cmd_2: KillCmdConfig = field(env='KILL')

        # `kill_cmd_3.signal` will be loaded from `SIGNAL`.
        kill_cmd_3: KillCmdConfig = field(env='')

You can also disable loading a field from an environment altogether:

.. code-block:: python

    class KillCmdConfig(Config):
        # Will not be loaded from env.
        pid: int = field(env=yuio.DISABLED)

To prefix all variable names with some string, pass the `prefix` parameter
to the :meth:`~Config.load_from_env` function:

.. code-block:: python

    # config.kill_cmd.field will be loaded
    # from `MY_APP_KILL_CMD_SIGNAL`
    config = BigConfig.load_from_env('MY_APP')


Parsing config files
--------------------

You can load config from structured config files,
such as `json`, `yaml` or `toml`:

.. skip: next

.. code-block:: python

    class ExecutorConfig(Config):
        threads: int
        use_gpu: bool = True

    class AppConfig(Config):
        executor: ExecutorConfig
        model: pathlib.Path

    config = AppConfig.load_from_json_file('~/.my_app_cfg.json')

In this example, contents of the above config would be:

.. code-block:: json

    {
        "executor": {
            "threads": 16,
            "use_gpu": true
        },
        "model": "/path/to/model"
    }

Note that, unlike with environment variables,
there is no way to inline nested configs.


Additional config validation
----------------------------

If you have invariants that can't be captured with type system,
you can override :meth:`~Config.validate_config`. This method will be called
every time you load a config from file, arguments or environment:

.. code-block:: python

    class DocGenConfig(Config):
        categories: list[str] = ["quickstart", "api_reference"]
        category_names: dict[str, str] = {"deep_dive": "Deep Dive"}

        def validate_config(self):
            for category in self.category_names:
                if category not in self.categories:
                    raise yuio.parse.ParsingError(f"unknown category {category}")


Merging configs
---------------

Configs are specially designed to be merge-able. The basic pattern is to create
an empty config instance, then :meth:`~Config.update` it with every config source:

.. skip: next

.. code-block:: python

    config = AppConfig()
    config.update(AppConfig.load_from_json_file("~/.my_app_cfg.json"))
    config.update(AppConfig.load_from_env())
    # ...and so on.

The :meth:`~Config.update` function ignores default values, and only overrides
keys that were actually configured.

If you need a more complex update behavior, you can add a merge function for a field:

.. code-block:: python

    class AppConfig(Config):
        plugins: list[str] = field(
            default=[],
            merge=lambda left, right: [*left, *right],
        )

Here, whenever we :meth:`~Config.update` ``AppConfig``, ``plugins`` from both instances
will be concatenated.

.. warning::

    Merge function shouldn't mutate its arguments.
    It should produce a new value instead.

.. warning::

    Merge function will not be called for default value. It's advisable to keep the
    default value empty, and add the actual default to the initial empty config:

    .. skip: next

    .. code-block:: python

        config = AppConfig(plugins=["markdown", "rst"])
        config.update(...)

"""

from __future__ import annotations

import argparse
import json
import os
import pathlib
import textwrap
from dataclasses import dataclass

import yuio
import yuio.complete
import yuio.json_schema
import yuio.parse
from yuio import _typing as _t

__all__ = [
    "MutuallyExclusiveGroup",
    "field",
    "inline",
    "positional",
    "Config",
]

T = _t.TypeVar("T")
Cfg = _t.TypeVar("Cfg", bound="Config")


class MutuallyExclusiveGroup:
    """
    A sentinel for creating mutually exclusive groups.

    """


@dataclass(frozen=True)
class _FieldSettings:
    default: _t.Any
    parser: yuio.parse.Parser[_t.Any] | None = None
    help: str | yuio.Disabled | None = None
    env: str | yuio.Disabled | None = None
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None
    completer: yuio.complete.Completer | None = None
    metavar: str | None = None
    required: bool = False
    merge: _t.Callable[[_t.Any, _t.Any], _t.Any] | None = None
    group: MutuallyExclusiveGroup | None = None
    usage: yuio.Omit | bool | None = None

    def _update_defaults(
        self,
        qualname: str,
        name: str,
        ty_with_extras: _t.Any,
        parsed_help: str | None,
        allow_positionals: bool,
    ) -> _Field:
        ty = ty_with_extras
        while _t.get_origin(ty) is _t.Annotated:
            ty = _t.get_args(ty)[0]
        is_subconfig = isinstance(ty, type) and issubclass(ty, Config)

        help: str | yuio.Disabled
        if self.help is not None:
            help = self.help
        elif parsed_help is not None:
            help = parsed_help
        elif is_subconfig and ty.__doc__:
            help = ty.__doc__
        else:
            help = ""
        if help == argparse.SUPPRESS:
            help = yuio.DISABLED

        env: str | yuio.Disabled
        if self.env is not None:
            env = self.env
        else:
            env = name.upper()
        if env == "" and not is_subconfig:
            raise TypeError(f"{qualname} got an empty env variable name")

        flags: list[str] | yuio.Positional | yuio.Disabled
        if self.flags is yuio.DISABLED or self.flags is yuio.POSITIONAL:
            flags = self.flags
            if not allow_positionals and flags is yuio.POSITIONAL:
                raise TypeError(
                    f"{qualname}: positional arguments are not allowed in configs"
                )
        elif self.flags is None:
            flags = ["--" + name.replace("_", "-")]
        elif isinstance(self.flags, str):
            flags = [self.flags]
        else:
            if not self.flags:
                raise TypeError(f"{qualname} should have at least one flag")
            flags = self.flags
        if flags is not yuio.DISABLED and flags is not yuio.POSITIONAL:
            for flag in flags:
                if flag and not flag.startswith("-"):
                    raise TypeError(f"{qualname}: flag should start with a dash")
                if not flag and not is_subconfig:
                    raise TypeError(f"{qualname} got an empty flag")

        default = self.default

        parser = self.parser

        required = self.required

        if is_subconfig:
            if default is not yuio.MISSING:
                raise TypeError(
                    f"error in {qualname}: nested configs can't have defaults"
                )

            if parser is not None:
                raise TypeError(
                    f"error in {qualname}: nested configs can't have parsers"
                )

            if flags is not yuio.DISABLED:
                if flags is yuio.POSITIONAL:
                    raise TypeError(
                        f"error in {qualname}: nested configs can't be positional"
                    )
                if len(flags) > 1:
                    raise TypeError(
                        f"error in {qualname}: nested configs can't have multiple flags"
                    )
                if flags[0] and not flags[0].startswith("--"):
                    raise TypeError(
                        f"error in {qualname}: nested configs can't have a short flag"
                    )
        elif parser is None:
            try:
                parser = yuio.parse.from_type_hint(ty_with_extras)
            except TypeError as e:
                raise TypeError(
                    f"can't derive parser for {qualname}:\n"
                    + textwrap.indent(str(e), "  ")
                ) from None

        if parser is not None:
            origin = _t.get_origin(ty)
            args = _t.get_args(ty)

            is_optional = (
                default is None
                or _t.is_union(origin)
                and len(args) == 2
                and type(None) in args
            )

            if is_optional and not yuio.parse._is_optional_parser(parser):
                parser = yuio.parse.Optional(parser)

            if (
                flags is yuio.POSITIONAL
                and default is not yuio.MISSING
                and parser.supports_parse_many()
            ):
                raise TypeError(
                    f"{qualname}: positional multi-value arguments can't have defaults"
                )

        completer = self.completer
        metavar = self.metavar
        if parser is not None and (completer is not None or metavar is not None):
            parser = yuio.parse.WithMeta(parser, desc=metavar, completer=completer)

        merge = self.merge

        if is_subconfig and merge is not None:
            raise TypeError(f"error in {qualname}: nested configs can't have merge")

        group = self.group

        if is_subconfig and group is not None:
            raise TypeError(
                f"error in {qualname}: nested configs can't be a part "
                "of a mutually exclusive group"
            )

        usage = self.usage

        return _Field(
            default,
            parser,
            help,
            env,
            flags,
            is_subconfig,
            ty,
            required,
            merge,
            group,
            usage,
        )


@dataclass(frozen=True)
class _Field:
    default: _t.Any
    parser: yuio.parse.Parser[_t.Any] | None
    help: str | yuio.Disabled
    env: str | yuio.Disabled
    flags: list[str] | yuio.Positional | yuio.Disabled
    is_subconfig: bool
    ty: type
    required: bool
    merge: _t.Callable[[_t.Any, _t.Any], _t.Any] | None
    group: MutuallyExclusiveGroup | None
    usage: yuio.Omit | bool | None = None


@_t.overload
def field(
    *,
    completer: yuio.complete.Completer | None = None,
    metavar: str | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None,
    group: MutuallyExclusiveGroup | None = None,
    usage: yuio.Omit | bool | None = None,
) -> _t.Any: ...


@_t.overload
def field(
    *,
    default: None,
    parser: yuio.parse.Parser[T] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None = None,
    metavar: str | None = None,
    merge: _t.Callable[[T, T], T] | None = None,
    group: MutuallyExclusiveGroup | None = None,
    usage: yuio.Omit | bool | None = None,
) -> T | None: ...


@_t.overload
def field(
    *,
    default: T | yuio.Missing = yuio.MISSING,
    parser: yuio.parse.Parser[T] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None = None,
    metavar: str | None = None,
    merge: _t.Callable[[T, T], T] | None = None,
    group: MutuallyExclusiveGroup | None = None,
    usage: yuio.Omit | bool | None = None,
) -> T: ...


def field(
    *,
    default: _t.Any = yuio.MISSING,
    parser: yuio.parse.Parser[_t.Any] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    flags: str | list[str] | yuio.Positional | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None = None,
    metavar: str | None = None,
    merge: _t.Callable[[_t.Any, _t.Any], _t.Any] | None = None,
    group: MutuallyExclusiveGroup | None = None,
    usage: yuio.Omit | bool | None = None,
) -> _t.Any:
    """
    Field descriptor, used for additional configuration of CLI arguments
    and config fields.

    In apps::

        @app
        def main(
            # Will be loaded from `--input`.
            input: pathlib.Path | None = None,

            # Will be loaded from `-o` or `--output`.
            output: pathlib.Path | None = field(default=None, flags=['-p', '--pid'])
        ):
            ...

    In configs::

        class AppConfig(Config):
            model: pathlib.Path | None = field(
                default=None,
                help="trained model to execute",
            )

    :param default:
        default value for the field or CLI argument.
    :param parser:
        parser that will be used to parse and CLI arguments.
    :param help:
        help message that will be used in CLI argument description.

        Pass :data:`~yuio.DISABLED` to remove this field from CLI help.

        By default, help message is inferred from comments right above the field
        definition (comments must start with ``#:``).

        Help messages are formatted using Markdown (see :mod:`yuio.md`).
    :param env:
        in configs, specifies name of environment variable that will be used
        if loading config from environment variable.

        Pass :data:`~yuio.DISABLED` to disable loading this field form environment variable.

        Pass an empty string to disable prefixing nested config variables.
    :param flags:
        list of names (or a single name) of CLI flags that will be used for this field.

        In configs, pass :data:`~yuio.DISABLED` to disable loading this field form CLI arguments.

        In apps, pass :data:`~yuio.POSITIONAL` to make this argument positional.

        Pass an empty string to disable prefixing nested config flags.
    :param completer:
        completer that will be used for autocompletion in CLI. Using this option
        is equivalent to overriding ``completer`` with :class:`yuio.parse.WithMeta`.
    :param metavar:
        value description that will be used for CLI help messages. Using this option
        is equivalent to overriding ``desc`` with :class:`yuio.parse.WithMeta`.
    :param merge:
        defines how values of this field are merged when configs are updated.
    :param group:
        defines mutually exclusive group for this field. Create an instance
        of :class:`yuio.app.MutuallyExclusiveGroup` and pass it to all fields that
        should be mutually exclusive.
    :param usage:
        controls how this field renders in CLI usage section. Passing :data:`False`
        removes this field from usage, and passing :class:`yuio.OMIT` replaces all
        omitted fields with a single string ``"<options>"``.

    """

    return _FieldSettings(
        default=default,
        completer=completer,
        metavar=metavar,
        parser=parser,
        help=help,
        env=env,
        flags=flags,
        merge=merge,
        group=group,
        usage=usage,
    )


def inline(
    help: str | yuio.Disabled | None = None,
    usage: yuio.Omit | bool | None = None,
) -> _t.Any:
    """
    A shortcut for inlining nested configs.

    Equivalent to calling :func:`~yuio.app.field` with ``env`` and ``flags``
    set to an empty string.

    """

    return field(help=help, env="", flags="", usage=usage)


@_t.overload
def positional(
    *,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None = None,
    metavar: str | None = None,
    usage: yuio.Omit | bool | None = None,
) -> _t.Any: ...


@_t.overload
def positional(
    *,
    default: None,
    parser: yuio.parse.Parser[T] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None = None,
    metavar: str | None = None,
    usage: yuio.Omit | bool | None = None,
) -> T | None: ...


@_t.overload
def positional(
    *,
    default: T | yuio.Missing = yuio.MISSING,
    parser: yuio.parse.Parser[T] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None = None,
    metavar: str | None = None,
    usage: yuio.Omit | bool | None = None,
) -> T: ...


def positional(
    *,
    default: _t.Any = yuio.MISSING,
    parser: yuio.parse.Parser[_t.Any] | None = None,
    help: str | yuio.Disabled | None = None,
    env: str | yuio.Disabled | None = None,
    completer: yuio.complete.Completer | None = None,
    metavar: str | None = None,
    usage: yuio.Omit | bool | None = None,
) -> _t.Any:
    """
    A shortcut for adding a positional argument.

    Equivalent to calling :func:`field` with ``flags`` set to :data:`~yuio.POSITIONAL`.

    """

    return field(
        default=default,
        parser=parser,
        help=help,
        env=env,
        flags=yuio.POSITIONAL,
        completer=completer,
        metavar=metavar,
        usage=usage,
    )


def _action(
    field: _Field,
    usage: yuio.Omit | bool,
    parse_many: bool = False,
    const: _t.Any = yuio.MISSING,
):
    class Action(argparse.Action):
        @staticmethod
        def get_parser():
            return field.parser

        @staticmethod
        def get_merge():
            return field.merge

        @staticmethod
        def get_usage():
            return usage

        def __call__(self, _, namespace, values, option_string=None):
            try:
                if const is not yuio.MISSING:
                    assert field.parser
                    assert values == []
                    parsed = field.parser.parse_config(const)
                elif parse_many:
                    if values is yuio.MISSING:
                        values = []
                    assert values is not None and not isinstance(values, str)
                    assert field.parser
                    parsed = field.parser.parse_many(values)
                else:
                    if values is yuio.MISSING:
                        return
                    assert isinstance(values, str)
                    assert field.parser
                    parsed = field.parser.parse(values)
            except argparse.ArgumentTypeError as e:
                raise argparse.ArgumentError(self, str(e))
            # Note: merge will be executed in `namespace.__setattr__`,
            # see `yuio.app._Namespace`.
            setattr(namespace, self.dest, parsed)

    return Action


@_t.dataclass_transform(
    eq_default=False,
    order_default=False,
    kw_only_default=True,
    frozen_default=False,
    field_specifiers=(field, inline, positional),
)
class Config:
    """
    Base class for configs.

    Pass keyword args to set fields, or pass another config to copy it::

        Config(config1, config2, ..., field1=value1, ...)

    Upon creation, all fields that aren't explicitly initialized
    and don't have defaults are considered missing.
    Accessing them will raise :class:`AttributeError`.

    .. automethod:: update

    .. automethod:: load_from_env

    .. automethod:: load_from_json_file

    .. automethod:: load_from_yaml_file

    .. automethod:: load_from_toml_file

    .. automethod:: load_from_parsed_file

    .. automethod:: to_json_schema

    .. automethod:: validate_config

    """

    @classmethod
    def __get_fields(cls) -> dict[str, _Field]:
        if cls.__fields is not None:
            return cls.__fields

        try:
            docs = yuio._find_docs(cls)
        except Exception:
            yuio._logger.exception(
                "unable to get documentation for class %s",
                cls.__qualname__,
            )
            docs = {}

        fields = {}
        defaults = {}

        for base in reversed(cls.__mro__):
            if base is not cls and hasattr(base, "_Config__get_fields"):
                fields.update(getattr(base, "_Config__get_fields")())

        try:
            types = _t.get_type_hints(cls, include_extras=True)
        except NameError as e:
            if "<locals>" in cls.__qualname__:
                raise NameError(
                    f"{e}. "
                    f"Note: forward references do not work inside functions "
                    f"(see https://github.com/python/typing/issues/797)"
                ) from None
            raise

        for name in cls.__annotations__:
            if name.startswith("_"):
                continue

            value = cls.__dict__.get(name, yuio.MISSING)
            if isinstance(value, _FieldSettings):
                field = value
            else:
                field = _FieldSettings(default=value)

            fields[name] = full_field = field._update_defaults(
                f"{cls.__qualname__}.{name}",
                name,
                types[name],
                docs.get(name),
                cls.__allow_positionals,
            )
            if full_field.is_subconfig:
                defaults[name] = full_field.ty
            else:
                defaults[name] = full_field.default

        # We don't want to set any attributes on cls if any `_update_defaults` has
        # raised an exception. For this reason, we defer setting defaults
        # until all fields were processed.
        for name, default in defaults.items():
            setattr(cls, name, default)
        cls.__fields = fields

        return fields

    def __init_subclass__(cls, _allow_positionals=None, **kwargs):
        super().__init_subclass__(**kwargs)

        if _allow_positionals is not None:
            cls.__allow_positionals: bool = _allow_positionals
        cls.__fields: dict[str, _Field] | None = None

    def __init__(self, *args: _t.Self | dict[str, _t.Any], **kwargs):
        for name, field in self.__get_fields().items():
            if field.is_subconfig:
                setattr(self, name, field.ty())

        for arg in args:
            self.update(arg)

        self.update(kwargs)

    def update(self, other: _t.Self | dict[str, _t.Any], /):
        """
        Update fields in this config with fields from another config.

        This function is similar to :meth:`dict.update`.

        Nested configs are updated recursively.

        :param other:
            data for update.

        """

        if not other:
            return

        if isinstance(other, Config):
            if (
                self.__class__ not in other.__class__.__mro__
                and other.__class__ not in self.__class__.__mro__
            ):
                raise TypeError("updating from an incompatible config")
            ns = other.__dict__
        elif isinstance(other, dict):
            ns = other
            for name in ns:
                if name not in self.__get_fields():
                    raise TypeError(f"unknown field: {name}")
        else:
            raise TypeError("expected a dict or a config class")

        for name, field in self.__get_fields().items():
            if name in ns:
                if field.is_subconfig:
                    getattr(self, name).update(ns[name])
                elif ns[name] is not yuio.MISSING:
                    if field.merge is not None and name in self.__dict__:
                        setattr(self, name, field.merge(getattr(self, name), ns[name]))
                    else:
                        setattr(self, name, ns[name])

    @classmethod
    def load_from_env(cls, prefix: str = "") -> _t.Self:
        """
        Load config from environment variables.

        :param prefix:
            if given, names of all environment variables will be prefixed with
            this string and an underscore.

        """

        try:
            result = cls.__load_from_env(prefix)
            result.validate_config()
            return result
        except yuio.parse.ParsingError as e:
            raise yuio.parse.ParsingError(
                f"failed to load config from environment variables:\n"
                + textwrap.indent(str(e), "  ")
            ) from None

    @classmethod
    def __load_from_env(cls, prefix: str = "") -> _t.Self:

        fields = {}

        for name, field in cls.__get_fields().items():
            if field.env is yuio.DISABLED:
                continue

            if prefix and field.env:
                env = f"{prefix}_{field.env}"
            else:
                env = f"{prefix}{field.env}"

            if field.is_subconfig:
                fields[name] = field.ty.load_from_env(prefix=env)
            elif env in os.environ:
                assert field.parser is not None
                fields[name] = field.parser.parse(os.environ[env])

        return cls(**fields)

    @classmethod
    def _load_from_namespace(
        cls,
        namespace: argparse.Namespace,
        /,
        *,
        ns_prefix: str = "",
    ) -> _t.Self:
        result = cls.__load_from_namespace(namespace, ns_prefix + ":")
        result.validate_config()
        return result

    @classmethod
    def __load_from_namespace(
        cls, namespace: argparse.Namespace, prefix: str
    ) -> _t.Self:
        fields = {}

        for name, field in cls.__get_fields().items():
            if field.flags is yuio.DISABLED:
                continue

            dest = prefix + name

            if field.is_subconfig:
                fields[name] = field.ty.__load_from_namespace(namespace, dest + ".")
            elif hasattr(namespace, dest):
                fields[name] = getattr(namespace, dest)

        return cls(**fields)

    @classmethod
    def _setup_arg_parser(
        cls,
        parser: argparse.ArgumentParser,
        /,
        *,
        group: argparse.ArgumentParser | None = None,
        ns_prefix: str = "",
    ):
        group = group or parser
        cls.__setup_arg_parser(group, parser, "", ns_prefix + ":", False, 0, True)

    @classmethod
    def __setup_arg_parser(
        cls,
        group: argparse.ArgumentParser,
        parser: argparse.ArgumentParser,
        prefix: str,
        dest_prefix: str,
        suppress_help: bool,
        depth: int,
        usage: yuio.Omit | bool,
    ):
        if prefix:
            prefix += "-"

        mutex_groups = {}

        for name, field in cls.__get_fields().items():
            if field.flags is yuio.DISABLED:
                continue

            dest = dest_prefix + name

            if suppress_help or field.help is yuio.DISABLED:
                help = argparse.SUPPRESS
                current_suppress_help = True
                current_usage = field.usage if field.usage is not None else False
            else:
                help = field.help
                current_suppress_help = False
                current_usage = field.usage if field.usage is not None else usage

            flags: list[str] | yuio.Positional
            if prefix and field.flags is not yuio.POSITIONAL:
                flags = [prefix + flag.lstrip("-") for flag in field.flags]
            else:
                flags = field.flags

            if field.is_subconfig:
                assert flags is not yuio.POSITIONAL
                if current_suppress_help:
                    subgroup = group
                else:
                    lines = help.split("\n\n", 1)
                    title = lines[0].replace("\n", " ").rstrip(".") or name
                    desc = textwrap.dedent(lines[1]) if len(lines) > 1 else None
                    subgroup = parser.add_argument_group(title, desc)
                field.ty.__setup_arg_parser(
                    subgroup,
                    parser,
                    flags[0],
                    dest + ".",
                    current_suppress_help,
                    depth + 1,
                    current_usage,
                )
                continue
            else:
                assert field.parser is not None

            parse_many = field.parser.supports_parse_many()

            if flags is yuio.POSITIONAL:
                metavar = f"<{name.replace('_', '-')}>"
            elif parse_many:
                metavar = field.parser.describe_many_or_def()
            else:
                metavar = field.parser.describe_or_def()

            nargs = field.parser.get_nargs()
            if (
                flags is yuio.POSITIONAL
                and field.default is not yuio.MISSING
                and nargs is None
            ):
                nargs = "?"
            nargs_kw: _t.Any = {"nargs": nargs} if nargs is not None else {}

            if field.group is not None:
                if field.group not in mutex_groups:
                    mutex_groups[field.group] = group.add_mutually_exclusive_group()
                field_group = mutex_groups[field.group]
            else:
                field_group = None

            if flags is yuio.POSITIONAL:
                (field_group or group).add_argument(
                    dest,
                    default=yuio.MISSING,
                    help=help,
                    metavar=metavar,
                    action=_action(field, current_usage, parse_many),
                    **nargs_kw,
                )
            elif yuio.parse._is_bool_parser(field.parser):
                mutex_group = field_group or group.add_mutually_exclusive_group(
                    required=field.required
                )

                if depth > 0 or not isinstance(field.default, bool):
                    pos_help = help
                    neg_help = None
                elif field.default:
                    pos_help = argparse.SUPPRESS
                    neg_help = help
                else:
                    pos_help = help
                    neg_help = argparse.SUPPRESS

                mutex_group.add_argument(
                    *flags,
                    default=yuio.MISSING,
                    help=pos_help,
                    dest=dest,
                    action=_action(field, current_usage, const=True),
                    nargs=0,
                )

                assert field.flags is not yuio.POSITIONAL
                for flag in field.flags:
                    if flag.startswith("--"):
                        flag_neg = (prefix or "--") + "no-" + flag[2:]
                        if current_suppress_help or neg_help == argparse.SUPPRESS:
                            help = argparse.SUPPRESS
                        elif neg_help:
                            help = neg_help
                        else:
                            help = f'disable <c hl/flag:sh-usage>{(prefix or "--") + flag[2:]}</c>'
                        mutex_group.add_argument(
                            flag_neg,
                            default=yuio.MISSING,
                            help=help,
                            dest=dest,
                            action=_action(field, current_usage, const=False),
                            nargs=0,
                        )
                        break
            else:
                (field_group or group).add_argument(
                    *flags,
                    default=yuio.MISSING,
                    help=help,
                    metavar=metavar,
                    required=field.required,
                    dest=dest,
                    action=_action(field, current_usage, parse_many),
                    **nargs_kw,
                )

    @classmethod
    def load_from_json_file(
        cls,
        path: str | pathlib.Path,
        /,
        *,
        ignore_unknown_fields: bool = False,
        ignore_missing_file: bool = False,
    ) -> _t.Self:
        """
        Load config from a ``.json`` file.

        :param path:
            path of the config file.
        :param ignore_unknown_fields:
            if :data:`True`, this method will ignore fields that aren't listed
            in config class.
        :param ignore_missing_file:
            if :data:`True`, silently ignore a file missing error. This is useful
            when loading a config from a home directory.

        """

        return cls.__load_from_file(
            path, json.loads, ignore_unknown_fields, ignore_missing_file
        )

    @classmethod
    def load_from_yaml_file(
        cls,
        path: str | pathlib.Path,
        /,
        *,
        ignore_unknown_fields: bool = False,
        ignore_missing_file: bool = False,
    ) -> _t.Self:
        """
        Load config from a ``.yaml`` file.

        This requires `PyYaml <https://pypi.org/project/PyYAML/>`_ package
        to be installed.

        :param path:
            path of the config file.
        :param ignore_unknown_fields:
            if :data:`True`, this method will ignore fields that aren't listed
            in config class.
        :param ignore_missing_file:
            if :data:`True`, silently ignore a file missing error. This is useful
            when loading a config from a home directory.

        """

        try:
            import yaml
        except ImportError:
            raise ImportError("PyYaml is not available")

        return cls.__load_from_file(
            path, yaml.safe_load, ignore_unknown_fields, ignore_missing_file
        )

    @classmethod
    def load_from_toml_file(
        cls,
        path: str | pathlib.Path,
        /,
        *,
        ignore_unknown_fields: bool = False,
        ignore_missing_file: bool = False,
    ) -> _t.Self:
        """
        Load config from a ``.toml`` file.

        This requires
        `tomllib <https://docs.python.org/3/library/tomllib.html>`_ or
        `toml <https://pypi.org/project/toml/>`_ package
        to be installed.

        :param path:
            path of the config file.
        :param ignore_unknown_fields:
            if :data:`True`, this method will ignore fields that aren't listed
            in config class.
        :param ignore_missing_file:
            if :data:`True`, silently ignore a file missing error. This is useful
            when loading a config from a home directory.

        """

        try:
            import toml
        except ImportError:
            try:
                import tomllib as toml  # type: ignore
            except ImportError:
                raise ImportError("toml is not available")

        return cls.__load_from_file(
            path, toml.loads, ignore_unknown_fields, ignore_missing_file
        )

    @classmethod
    def __load_from_file(
        cls,
        path: str | pathlib.Path,
        file_parser: _t.Callable[[str], _t.Any],
        ignore_unknown_fields: bool = False,
        ignore_missing_file: bool = False,
    ) -> _t.Self:
        if ignore_missing_file and not os.path.exists(path):
            return cls()

        try:
            with open(path) as file:
                loaded = file_parser(file.read())
        except Exception as e:
            raise yuio.parse.ParsingError(
                f"invalid config {path}:\n" + textwrap.indent(str(e), "  ")
            ) from None

        return cls.load_from_parsed_file(
            loaded, ignore_unknown_fields=ignore_unknown_fields, path=path
        )

    @classmethod
    def load_from_parsed_file(
        cls,
        parsed: dict[str, object],
        /,
        *,
        ignore_unknown_fields: bool = False,
        path: str | pathlib.Path | None = None,
    ) -> _t.Self:
        """
        Load config from parsed config file.

        This method takes a dict with arbitrary values that you'd get from
        parsing type-rich configs such as ``yaml`` or ``json``.

        For example::

            with open('conf.yaml') as file:
                config = Config.load_from_parsed_file(yaml.load(file))

        :param parsed:
            data from parsed file.
        :param ignore_unknown_fields:
            if :data:`True`, this method will ignore fields that aren't listed
            in config class.
        :param path:
            path of the original file, used for error reporting.

        """

        try:
            result = cls.__load_from_parsed_file(parsed, ignore_unknown_fields, "")
            result.validate_config()
            return result
        except yuio.parse.ParsingError as e:
            if path is None:
                raise
            else:
                raise yuio.parse.ParsingError(
                    f"invalid config {path}:\n" + textwrap.indent(str(e), "  ")
                ) from None

    @classmethod
    def __load_from_parsed_file(
        cls,
        parsed: dict[str, object],
        ignore_unknown_fields: bool = False,
        field_prefix: str = "",
    ) -> _t.Self:
        if not isinstance(parsed, dict):
            raise TypeError("config should be a dict")

        fields = {}

        if not ignore_unknown_fields:
            for name in parsed:
                if name not in cls.__get_fields() and name != "$schema":
                    raise yuio.parse.ParsingError(f"unknown field {field_prefix}{name}")

        for name, field in cls.__get_fields().items():
            if name in parsed:
                if field.is_subconfig:
                    fields[name] = field.ty.__load_from_parsed_file(
                        parsed[name], ignore_unknown_fields, field_prefix=name + "."
                    )
                else:
                    assert field.parser is not None
                    try:
                        value = field.parser.parse_config(parsed[name])
                    except yuio.parse.ParsingError as e:
                        raise yuio.parse.ParsingError(
                            f"can't parse {field_prefix}{name}:\n"
                            + textwrap.indent(str(e), "  ")
                        ) from None
                    fields[name] = value

        return cls(**fields)

    def __getattribute(self, item):
        value = super().__getattribute__(item)
        if value is yuio.MISSING:
            raise AttributeError(f"{item} is not configured")
        else:
            return value

    # A dirty hack to hide `__getattribute__` from type checkers.
    locals()["__getattribute__"] = __getattribute

    def __repr__(self):
        return self.__repr(0)

    def __repr(self, indent):
        field_reprs = []
        prefix = " " * indent
        for name in self.__get_fields():
            value = getattr(self, name, yuio.MISSING)
            if isinstance(value, Config):
                value_repr = value.__repr(indent + 2)
            else:
                value_repr = repr(value)
            field_reprs.append(f"{prefix}  {name}={value_repr}")
        if field_reprs:
            field_desc = ",\n".join(field_reprs)
            return f"{self.__class__.__name__}(\n{field_desc}\n{prefix})"
        else:
            return f"{self.__class__.__name__}()"

    def validate_config(self):
        """
        Perform config validation.

        This function is called every time a config is loaded from CLI arguments,
        file, or environment variables. It should check that config is correct,
        and raise :class:`yuio.parse.ParsingError` if it's not.

        """

    @classmethod
    def to_json_schema(
        cls, ctx: yuio.json_schema.JsonSchemaContext
    ) -> yuio.json_schema.JsonSchemaType:
        """
        Create a Json schema object based on this config.

        The purpose of this method is to make schemas for use in IDEs, i.e. to provide
        autocompletion or simple error checking. The returned schema is not guaranteed
        to reflect all constraints added to the parser.

        :param ctx:
            context for building a schema.

        """

        return ctx.add_type(cls, _t.type_repr(cls), lambda: cls.__to_json_schema(ctx))

    @classmethod
    def __to_json_schema(
        cls, ctx: yuio.json_schema.JsonSchemaContext
    ) -> yuio.json_schema.JsonSchemaType:
        properties: dict[str, yuio.json_schema.JsonSchemaType] = {}
        defaults = {}

        properties["$schema"] = yuio.json_schema.String()

        for name, field in cls.__get_fields().items():
            if field.is_subconfig:
                properties[name] = field.ty.to_json_schema(ctx)
            else:
                assert field.parser
                field_schema = field.parser.to_json_schema(ctx)
                if field.help and field.help is not yuio.DISABLED:
                    field_schema = yuio.json_schema.Meta(
                        field_schema, description=field.help
                    )
                properties[name] = field_schema
                if field.default is not yuio.MISSING:
                    try:
                        defaults[name] = field.parser.to_json_value(field.default)
                    except TypeError:
                        pass

        return yuio.json_schema.Meta(
            yuio.json_schema.Object(properties),
            title=cls.__name__,
            description=cls.__doc__,
            default=defaults,
        )


Config.__init_subclass__(_allow_positionals=False)



import yuio.io
import yuio.widget

if __name__ == "__main__":
    term = yuio.io.get_term()
    theme = yuio.io.get_theme()

    options = [
        yuio.widget.Option(display_text="GameBoy", value="GameBoy"),
        yuio.widget.Option(display_text="Wii", value="Wii"),
        yuio.widget.Option(display_text="PlayStation2", value="PlayStation2"),
        yuio.widget.Option(
            display_text="PlayStationPortable", value="PlayStationPortable"
        ),
        yuio.widget.Option(display_text="XboxOne", value="XboxOne"),
        yuio.widget.Option(display_text="MegaDrive/Genesis", value="MegaDrive/Genesis"),
        yuio.widget.Option(display_text="Atari2600", value="Atari2600"),
    ]

    widget = yuio.widget.Multiselect(options)

    yuio.io.heading("Choose consoles that you like")
    result = widget.run(term, theme)
    yuio.io.info("You selected `%s`", result)



# import pytest

# from yuio import _typing as _t
# from yuio.app import *
# from yuio.config import *


# class TestArgs:
#     def test_no_args(self):
#         pass


# class TestPositionals:
#     def test_single_positional(self, capsys):
#         run_result = {}

#         @app
#         def cmd(p: int = positional()):
#             run_result.update(locals())

#         try:
#             cmd.run("10".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == 10

#         with pytest.raises(SystemExit):
#             cmd.run("".split())
#         assert "the following arguments are required: <p>" in capsys.readouterr().err

#         with pytest.raises(SystemExit):
#             cmd.run("1 2".split())
#         assert "unrecognized arguments: 2" in capsys.readouterr().err

#         with pytest.raises(SystemExit):
#             cmd.run("abc".split())
#         assert "can't parse 'abc' as an int" in capsys.readouterr().err

#     def test_single_positional_with_default(self, capsys):
#         run_result = {}

#         @app
#         def cmd(p: int = positional(default=15)):
#             run_result.update(locals())

#         try:
#             cmd.run("10".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == 10

#         try:
#             cmd.run("".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == 15

#         with pytest.raises(SystemExit):
#             cmd.run("abc".split())
#         assert "can't parse 'abc' as an int" in capsys.readouterr().err

#     def test_optional_positional(self, capsys):
#         run_result = {}

#         @app
#         def cmd(p: _t.Optional[int] = positional()):
#             run_result.update(locals())

#         try:
#             cmd.run("10".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == 10

#         with pytest.raises(SystemExit):
#             cmd.run("".split())
#         assert "the following arguments are required: <p>" in capsys.readouterr().err

#         with pytest.raises(SystemExit):
#             cmd.run("abc".split())
#         assert "can't parse 'abc' as an int" in capsys.readouterr().err

#     def test_optional_positional_with_default(self, capsys):
#         run_result = {}

#         @app
#         def cmd(p: _t.Optional[int] = positional(default=15)):
#             run_result.update(locals())

#         try:
#             cmd.run("10".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == 10

#         try:
#             cmd.run("".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == 15

#         with pytest.raises(SystemExit):
#             cmd.run("abc".split())
#         assert "can't parse 'abc' as an int" in capsys.readouterr().err

#     def test_list_positional(self, capsys):
#         run_result = {}

#         @app
#         def cmd(p: _t.List[int] = positional()):
#             run_result.update(locals())

#         try:
#             cmd.run("".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == []

#         try:
#             cmd.run("10".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == [10]

#         try:
#             cmd.run("10 11 12".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == [10, 11, 12]

#         with pytest.raises(SystemExit):
#             cmd.run("10 abc 12".split())
#         assert "can't parse 'abc' as an int" in capsys.readouterr().err

#     def test_list_positional_len_bounds_ge(self, capsys):
#         run_result = {}

#         @app
#         def cmd(
#             p: _t.List[int] = positional(
#                 parser=yuio.parse.List(yuio.parse.Int()).len_ge(2)
#             ),
#         ):
#             run_result.update(locals())

#         try:
#             cmd.run("1 2".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == [1, 2]

#         try:
#             cmd.run("1 2 3".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == [1, 2, 3]

#         with pytest.raises(SystemExit):
#             cmd.run("".split())
#         assert "the following arguments are required: <p>" in capsys.readouterr().err

#         with pytest.raises(SystemExit):
#             cmd.run("1".split())
#         assert (
#             "argument <p>: length of a value should be greater or equal to 2, "
#             "got [1] instead" in capsys.readouterr().err
#         )

#     def test_list_positional_len_bounds_eq(self, capsys):
#         run_result = {}

#         @app
#         def cmd(
#             p: _t.List[int] = positional(
#                 parser=yuio.parse.List(yuio.parse.Int()).len_eq(2)
#             ),
#         ):
#             run_result.update(locals())

#         try:
#             cmd.run("1 2".split())
#         except SystemExit as e:
#             assert e.code == 0
#         assert run_result["p"] == [1, 2]

#         with pytest.raises(SystemExit):
#             cmd.run("".split())
#         assert "the following arguments are required: <p>" in capsys.readouterr().err

#         with pytest.raises(SystemExit):
#             cmd.run("1".split())
#         assert "the following arguments are required: <p>" in capsys.readouterr().err

#         with pytest.raises(SystemExit):
#             cmd.run("1 2 3".split())
#         assert "unrecognized arguments: 3" in capsys.readouterr().err

#     def test_list_positional_with_default(self):
#         @app
#         def cmd(p: _t.List[int] = positional(default=[])):
#             assert False, "should never launch"

#         with pytest.raises(
#             TypeError, match="positional multi-value arguments can't have defaults"
#         ):
#             cmd.run("".split())

#     def test_tuple_positional(self):
#         pass

#     def test_tuple_positional_with_default(self):
#         pass


# class TestSubCommands:
#     # test subcommands with positionals
#     # test context
#     pass


# class TestSubConfigs:
#     pass


# class TestHelp:
#     pass


# class TestCompletions:
#     pass



import yuio.app
import yuio.io

@yuio.app.app
def main(greeting: str = "world"):
    yuio.io.md("""
        # Dear %s,

        - You can use `` `backticks` `` with all functions from `yuio.io`.
        - You can also use CommonMark block markup with `yuio.io.md`.
        - For example, check out this fork bomb:

          ```sh
          :(){ :|:& };:  # <- don't paste this in bash!
          ```
    """, greeting)

if __name__ == "__main__":
    main.run()



import yuio.app
import yuio.io

@yuio.app.app
def main():
    greeting = yuio.io.edit(
        "# Please, edit the greeting:\n"
        "Hello, world!",
        comment_marker="#"
    )

    yuio.io.info("Greeting: `%s`", greeting)

if __name__ == "__main__":
    main.run()



import logging
import os
import pathlib

import pytest

import yuio.exec


def test_exec():
    result = yuio.exec.exec("echo", "hello")
    assert result == "hello\n"


def test_input():
    result = yuio.exec.exec("cat", input="hello\n")
    assert result == "hello\n"


@pytest.mark.skipif(os.name == "nt", reason="bash is not available on windows")
def test_logging(capsys):
    logging.basicConfig(level=logging.DEBUG, force=True)
    logging.root.manager._clear_cache()  # type: ignore

    result = yuio.exec.exec("bash", "-c", "echo 'message' 1>&2; echo 'result'")
    assert result == "result\n"

    result = yuio.exec.exec(
        "bash", "-c", "echo 'message2' 1>&2; echo 'result2'", level=logging.INFO
    )
    assert result == "result2\n"

    res = capsys.readouterr()
    assert res.out == ""
    assert "DEBUG:yuio.exec:bash -c echo 'message' 1>&2; echo 'result'" in res.err
    assert "DEBUG:yuio.exec:message" in res.err
    assert "INFO:yuio.exec:bash -c echo 'message2' 1>&2; echo 'result2'" in res.err
    assert "INFO:yuio.exec:message2" in res.err


def test_env():
    result = yuio.exec.exec("env", env={"FOO": "BAR"})
    assert "FOO=BAR" in result


@pytest.mark.skipif(os.name == "nt", reason="bash is not available on windows")
def test_cwd(tmp_path):
    result = yuio.exec.exec("pwd", cwd=tmp_path)
    assert result.strip() == str(tmp_path)


def test_path(tmp_path):
    yuio.exec.exec("ls", pathlib.Path(tmp_path))


@pytest.mark.skipif(os.name == "nt", reason="bash is not available on windows")
def test_fail():
    with pytest.raises(yuio.exec.ExecError, match="message"):
        yuio.exec.exec("bash", "-c", "echo 'message' 1>&2; echo 'result'; exit 1")


@pytest.mark.skipif(os.name == "nt", reason="bash is not available on windows")
def test_sh():
    result = yuio.exec.sh("echo hello")
    assert result == "hello\n"



import pathlib
import yuio.app
import yuio.io

@yuio.app.app
def main(
    greeting: str = yuio.app.field(default="world", flags=["-g", "--greeting"]),
    output: pathlib.Path | None = yuio.app.positional(default=None),
):
    if output:
        output.write_text(f"Hello, {greeting}!\n")
    else:
        yuio.io.info(f"Hello, `%s`!", greeting)

if __name__ == "__main__":
    main.run()



from typing import Annotated
import yuio.app
import yuio.io
import yuio.parse

@yuio.app.app
def main(data: Annotated[list[int], yuio.parse.Json()]):
    yuio.io.success("Loaded data: `%r`", data)

if __name__ == "__main__":
    main.run()
